(* file kernel/ocaml/ml/numerix.ml: Objective-Caml extensible integer definitions
 +-----------------------------------------------------------------------+
 |  Copyright 2005-2006, Michel Quercia (michel.quercia@prepas.org)      |
 |                                                                       |
 |  This file is part of Numerix. Numerix is free software; you can      |
 |  redistribute it and/or modify it under the terms of the GNU Lesser   |
 |  General Public License as published by the Free Software Foundation; |
 |  either version 2.1 of the License, or (at your option) any later     |
 |  version.                                                             |
 |                                                                       |
 |  The Numerix Library is distributed in the hope that it will be       |
 |  useful, but WITHOUT ANY WARRANTY; without even the implied warranty  |
 |  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  |
 |  Lesser General Public License for more details.                      |
 |                                                                       |
 |  You should have received a copy of the GNU Lesser General Public     |
 |  License along with the GNU MP Library; see the file COPYING. If not, |
 |  write to the Free Software Foundation, Inc., 59 Temple Place -       |
 |  Suite 330, Boston, MA 02111-1307, USA.                               |
 +-----------------------------------------------------------------------+
 |                                                                       |
 |                           Dfinitions                                 |
 |                                                                       |
 +-----------------------------------------------------------------------*)

(* mode d'arrondi *)
type round_mode = Floor | Nearest_up | Ceil | Nearest_down

(* rsultat ternaire *)
type tristate = False | Unknown | True

module type Int_type = sig

  type t                     (* entier         *)
  type tref                  (* entier mutable *)

  val name : unit -> string  (* nom du module  *)

  val zero : t               (* the number 0   *)
  val one  : t               (* the number 1   *)

  (* rfrence      ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val make_ref    :                               t ->               tref   
  val copy_in     :               tref ->         t ->               unit   
  val copy_out    :               tref ->                            t      
  val look        :               tref ->                            t      
  
  (* addition       ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val add         :                               t -> t   ->        t      
  val add_1       :                               t -> int ->        t      
  val add_in      :               tref ->         t -> t   ->        unit   
  val add_1_in    :               tref ->         t -> int ->        unit   
  
  (* soustraction   ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val sub         :                               t -> t   ->        t      
  val sub_1       :                               t -> int ->        t      
  val sub_in      :               tref ->         t -> t   ->        unit   
  val sub_1_in    :               tref ->         t -> int ->        unit   
  
  (* multiplication ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val mul         :                               t -> t   ->        t      
  val mul_1       :                               t -> int ->        t      
  val mul_in      :               tref ->         t -> t   ->        unit   
  val mul_1_in    :               tref ->         t -> int ->        unit   
  
  (* division       ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val quomod      :                               t -> t   ->        t*t    
  val quo         :                               t -> t   ->        t      
  val modulo      :                               t -> t   ->        t      
  val gquomod     : round_mode ->                 t -> t   ->        t*t    
  val gquo        : round_mode ->                 t -> t   ->        t      
  val gmod        : round_mode ->                 t -> t   ->        t      
  
  val quomod_in   :               tref -> tref -> t -> t   ->        unit   
  val quo_in      :               tref ->         t -> t   ->        unit   
  val mod_in      :                       tref -> t -> t   ->        unit   
  val gquomod_in  : round_mode -> tref -> tref -> t -> t   ->        unit   
  val gquo_in     : round_mode -> tref ->         t -> t   ->        unit   
  val gmod_in     : round_mode ->         tref -> t -> t   ->        unit   
  
  val quomod_1    :                               t -> int ->        t*int  
  val quo_1       :                               t -> int ->        t      
  val mod_1       :                               t -> int ->        int    
  val gquomod_1   : round_mode ->                 t -> int ->        t*int  
  val gquo_1      : round_mode ->                 t -> int ->        t      
  val gmod_1      : round_mode ->                 t -> int ->        int    
  
  val quomod_1_in :               tref ->         t -> int ->        int    
  val quo_1_in    :               tref ->         t -> int ->        unit   
  val gquomod_1_in: round_mode -> tref ->         t -> int ->        int    
  val gquo_1_in   : round_mode -> tref ->         t -> int ->        unit   
  
  (* valeur absolue ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val abs         :                               t ->               t      
  val abs_in      :               tref ->         t ->               unit   
  
  (* oppos         ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val neg         :                               t ->               t      
  val neg_in      :               tref ->         t ->               unit   
  
  (* puissance p-me----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val sqr         :                               t ->               t      
  val pow         :                               t -> int ->        t      
  val pow_1       :                             int -> int ->        t      
  val powmod      :                               t -> t   -> t ->   t      
  val gpowmod     : round_mode ->                 t -> t   -> t ->   t      
  val sqr_in      :               tref ->         t ->               unit   
  val pow_in      :               tref ->         t -> int ->        unit   
  val pow_1_in    :               tref ->       int -> int ->        unit   
  val powmod_in   :               tref ->         t -> t   -> t ->   unit   
  val gpowmod_in  : round_mode -> tref ->         t -> t   -> t ->   unit   
  
  (* racine p-me    ----------------------------------------------------- *)
  (*                 mode          r       s       a    b      c      res  *)
  val sqrt        :                               t ->               t      
  val root        :                               t -> int ->        t      
  val gsqrt       : round_mode ->                 t ->               t      
  val groot       : round_mode ->                 t -> int ->        t      
  val sqrt_in     :               tref ->         t ->               unit   
  val root_in     :               tref ->         t -> int ->        unit   
  val gsqrt_in    : round_mode -> tref ->         t ->               unit   
  val groot_in    : round_mode -> tref ->         t -> int ->        unit   
  
  (* factorielle    ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val fact        :                             int ->               t      
  val fact_in     :               tref ->       int ->               unit   
  
  (* pgcd           ----------------------------------------------------- *)
  (*                  d     u     v     p    q    a    b      c      res  *)
  val gcd         :                               t -> t   ->        t      
  val gcd_ex      :                               t -> t   ->        t*t*t  
  val cfrac       :                               t -> t   ->     t*t*t*t*t 
  val gcd_in      : tref->                        t -> t   ->        unit   
  val gcd_ex_in   : tref->tref->tref->            t -> t   ->        unit   
  val cfrac_in    : tref->tref->tref->tref->tref->t -> t   ->        unit   
  
  (* primalit      ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val isprime     :                               t ->               tristate
  val isprime_1   :                             int ->               tristate

  (* comparaison    ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val sgn         :                               t ->               int    
  val cmp         :                               t -> t   ->        int    
  val cmp_1       :                               t -> int ->        int    
  val eq          :                               t -> t   ->        bool   
  val eq_1        :                               t -> int ->        bool   
  val neq         :                               t -> t   ->        bool   
  val neq_1       :                               t -> int ->        bool   
  val inf         :                               t -> t   ->        bool   
  val inf_1       :                               t -> int ->        bool   
  val infeq       :                               t -> t   ->        bool   
  val infeq_1     :                               t -> int ->        bool   
  val sup         :                               t -> t   ->        bool   
  val sup_1       :                               t -> int ->        bool   
  val supeq       :                               t -> t   ->        bool   
  val supeq_1     :                               t -> int ->        bool   
  
  (* conversion     ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val of_int      :                             int ->               t      
  val of_string   :                          string ->               t      
  val of_int_in   :               tref ->       int ->               unit   
  val of_string_in:               tref ->    string ->               unit   
  val int_of      :                               t ->               int    
  val string_of   :                               t ->               string 
  val bstring_of  :                               t ->               string 
  val hstring_of  :                               t ->               string 
  val ostring_of  :                               t ->               string 
  
  (* nombre alatoire --------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val nrandom     :                             int->                t      
  val zrandom     :                             int->                t      
  val nrandom1    :                             int->                t      
  val zrandom1    :                             int->                t      
  val nrandom_in  :               tref ->       int->                unit   
  val zrandom_in  :               tref ->       int->                unit   
  val nrandom1_in :               tref ->       int->                unit   
  val zrandom1_in :               tref ->       int->                unit   
  val random_init :                             int->                unit   
  
  (* reprsentation binaire --------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val nbits       :                               t ->               int    
  val lowbits     :                               t ->               int    
  val highbits    :                               t ->               int    
  val nth_word    :                               t -> int ->        int    
  val nth_bit     :                               t -> int ->        bool   
  
  (* dcalage       ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val shl         :                               t -> int ->        t      
  val shr         :                               t -> int ->        t      
  val split       :                               t -> int ->        t*t    
  val join        :                               t -> t   -> int -> t      
  val shl_in      :               tref ->         t -> int ->        unit   
  val shr_in      :               tref ->         t -> int ->        unit   
  val split_in    :               tref -> tref -> t -> int ->        unit   
  val join_in     :               tref ->         t -> t   -> int -> unit   
  
  (* affichage      ----------------------------------------------------- *)
  (*                mode          r       s       a    b      c      res  *)
  val toplevel_print   :                               t ->               unit
  val toplevel_print_tref:             tref ->                            unit
  
  (* exceptions *)
  exception Error of string

end

(* notation infixe *)
module Infixes(E : Int_type) = struct
  open E

  let (  ++  ) = add      
  let (  --  ) = sub      
  let (  **  ) = mul      
  let (  //  ) = quo      
  let (  %%  ) = modulo   
  let (  /%  ) = quomod   
  let (  <<  ) = shl      
  let (  >>  ) = shr      
  let (  ^^  ) = pow      

  let (  +=  ) r x = add_in r (look r) x
  let (  -=  ) r x = sub_in r (look r) x
  let (  *=  ) r x = mul_in r (look r) x
  let (  /=  ) r x = quo_in r (look r) x
  let (  %=  ) r x = mod_in r (look r) x
                
  let (  +.  ) = add_1    
  let (  -.  ) = sub_1    
  let (  *.  ) = mul_1    
  let (  /.  ) = quo_1    
  let (  %.  ) = mod_1    
  let (  /%. ) = quomod_1
  let (  ^.  ) = pow_1      
                
  let (  +=. ) r x = add_1_in r (look r) x
  let (  -=. ) r x = sub_1_in r (look r) x
  let (  *=. ) r x = mul_1_in r (look r) x
  let (  /=. ) r x = quo_1_in r (look r) x

  let (   =. ) = eq_1     
  let (  <>. ) = neq_1    
  let (   <. ) = inf_1    
  let (  <=. ) = infeq_1  
  let (   >. ) = sup_1    
  let (  >=. ) = supeq_1  

  let (  ~~  ) = look     

end

(* affiche en secondes le temps du processus, la diffrence avec le temps *)
(* prcdent et la chaine passe en argument.                             *)
external chrono : string -> unit = "chrono"


                           (* +-----------------+
                              |  module Clong   |
                              +-----------------+ *)

module Clong = struct

  type t
  type tref

  let name() = String.copy "Clong"

  external make_ref    :                               t ->               tref   = "cx_make_ref"
  external copy_in     :               tref ->         t ->               unit   = "cx_copy"
  external copy_out    :               tref ->                            t      = "cx_copy_out"
  external look        :               tref ->                            t      = "cx_look"
  external add         :                               t -> t   ->        t      = "cx_f_add"
  external add_1       :                               t -> int ->        t      = "cx_f_add_1"
  external add_in      :               tref ->         t -> t   ->        unit   = "cx_add"
  external add_1_in    :               tref ->         t -> int ->        unit   = "cx_add_1"
  external sub         :                               t -> t   ->        t      = "cx_f_sub"
  external sub_1       :                               t -> int ->        t      = "cx_f_sub_1"
  external sub_in      :               tref ->         t -> t   ->        unit   = "cx_sub"
  external sub_1_in    :               tref ->         t -> int ->        unit   = "cx_sub_1"
  external mul         :                               t -> t   ->        t      = "cx_f_mul"
  external mul_1       :                               t -> int ->        t      = "cx_f_mul_1"
  external mul_in      :               tref ->         t -> t   ->        unit   = "cx_mul"
  external mul_1_in    :               tref ->         t -> int ->        unit   = "cx_mul_1"
  external quomod      :                               t -> t   ->        t*t    = "cx_f_quomod"
  external quo         :                               t -> t   ->        t      = "cx_f_quo"
  external modulo      :                               t -> t   ->        t      = "cx_f_mod"
  external gquomod     : round_mode ->                 t -> t   ->        t*t    = "cx_f_gquomod"
  external gquo        : round_mode ->                 t -> t   ->        t      = "cx_f_gquo"
  external gmod        : round_mode ->                 t -> t   ->        t      = "cx_f_gmod"
  external quomod_in   :               tref -> tref -> t -> t   ->        unit   = "cx_quomod"
  external quo_in      :               tref ->         t -> t   ->        unit   = "cx_quo"
  external mod_in      :                       tref -> t -> t   ->        unit   = "cx_mod"
  external gquomod_in  : round_mode -> tref -> tref -> t -> t   ->        unit   = "cx_gquomod"
  external gquo_in     : round_mode -> tref ->         t -> t   ->        unit   = "cx_gquo"
  external gmod_in     : round_mode ->         tref -> t -> t   ->        unit   = "cx_gmod"
  external quomod_1    :                               t -> int ->        t*int  = "cx_f_quomod_1"
  external quo_1       :                               t -> int ->        t      = "cx_f_quo_1"
  external mod_1       :                               t -> int ->        int    = "cx_f_mod_1"
  external gquomod_1   : round_mode ->                 t -> int ->        t*int  = "cx_f_gquomod_1"
  external gquo_1      : round_mode ->                 t -> int ->        t      = "cx_f_gquo_1"
  external gmod_1      : round_mode ->                 t -> int ->        int    = "cx_f_gmod_1"
  external quomod_1_in :               tref ->         t -> int ->        int    = "cx_quomod_1"
  external quo_1_in    :               tref ->         t -> int ->        unit   = "cx_quo_1"
  external gquomod_1_in: round_mode -> tref ->         t -> int ->        int    = "cx_gquomod_1"
  external gquo_1_in   : round_mode -> tref ->         t -> int ->        unit   = "cx_gquo_1"
  external abs         :                               t ->               t      = "cx_f_abs"
  external abs_in      :               tref ->         t ->               unit   = "cx_abs"
  external neg         :                               t ->               t      = "cx_f_neg"
  external neg_in      :               tref ->         t ->               unit   = "cx_neg"
  external sqr         :                               t ->               t      = "cx_f_sqr"
  external pow         :                               t -> int ->        t      = "cx_f_pow"
  external pow_1       :                             int -> int ->        t      = "cx_f_pow_1"
  external powmod      :                               t -> t   -> t ->   t      = "cx_f_powmod"
  external gpowmod     : round_mode ->                 t -> t   -> t ->   t      = "cx_f_gpowmod"
  external sqr_in      :               tref ->         t ->               unit   = "cx_sqr"
  external pow_in      :               tref ->         t -> int ->        unit   = "cx_pow"
  external pow_1_in    :               tref ->       int -> int ->        unit   = "cx_pow_1"
  external powmod_in   :               tref ->         t -> t   -> t ->   unit   = "cx_powmod"
  external gpowmod_in  : round_mode -> tref ->         t -> t   -> t ->   unit   = "cx_gpowmod"
  external sqrt        :                               t ->               t      = "cx_f_sqrt"
  external root        :                               t -> int ->        t      = "cx_f_root"
  external gsqrt       : round_mode ->                 t ->               t      = "cx_f_gsqrt"
  external groot       : round_mode ->                 t -> int ->        t      = "cx_f_groot"
  external sqrt_in     :               tref ->         t ->               unit   = "cx_sqrt"
  external root_in     :               tref ->         t -> int ->        unit   = "cx_root"
  external gsqrt_in    : round_mode -> tref ->         t ->               unit   = "cx_gsqrt"
  external groot_in    : round_mode -> tref ->         t -> int ->        unit   = "cx_groot"
  external fact        :                             int ->               t      = "cx_f_fact"
  external fact_in     :               tref ->       int ->               unit   = "cx_fact"
  external gcd         :                               t -> t   ->        t      = "cx_f_gcd"
  external gcd_ex      :                               t -> t   ->        t*t*t  = "cx_f_gcd_ex"
  external cfrac       :                               t -> t   ->     t*t*t*t*t = "cx_f_cfrac"
  external gcd_in      : tref->                        t -> t   ->        unit   = "cx_gcd"
  external gcd_ex_in   : tref->tref->tref->            t -> t   ->        unit   = "cx_gcd_ex"
  external cfrac_in    : tref->tref->tref->tref->tref->t -> t   ->        unit   = "cx_cfrac_bytecode" "cx_cfrac"
  external isprime     :                               t ->           tristate   = "cx_isprime"
  external isprime_1   :                             int ->           tristate   = "cx_isprime_1"
  external sgn         :                               t ->               int    = "cx_sgn"
  external cmp         :                               t -> t   ->        int    = "cx_cmp"
  external cmp_1       :                               t -> int ->        int    = "cx_cmp_1"
  external eq          :                               t -> t   ->        bool   = "cx_eq"
  external eq_1        :                               t -> int ->        bool   = "cx_eq_1"
  external neq         :                               t -> t   ->        bool   = "cx_neq"
  external neq_1       :                               t -> int ->        bool   = "cx_neq_1"
  external inf         :                               t -> t   ->        bool   = "cx_inf"
  external inf_1       :                               t -> int ->        bool   = "cx_inf_1"
  external infeq       :                               t -> t   ->        bool   = "cx_infeq"
  external infeq_1     :                               t -> int ->        bool   = "cx_infeq_1"
  external sup         :                               t -> t   ->        bool   = "cx_sup"
  external sup_1       :                               t -> int ->        bool   = "cx_sup_1"
  external supeq       :                               t -> t   ->        bool   = "cx_supeq"
  external supeq_1     :                               t -> int ->        bool   = "cx_supeq_1"
  external of_int      :                             int ->               t      = "cx_of_int"
  external of_string   :                          string ->               t      = "cx_of_string"
  external of_int_in   :               tref ->       int ->               unit   = "cx_copy_int"
  external of_string_in:               tref ->    string ->               unit   = "cx_copy_string"
  external int_of      :                               t ->               int    = "cx_int_of"
  external string_of   :                               t ->               string = "cx_string_of"
  external bstring_of  :                               t ->               string = "cx_bstring_of"
  external hstring_of  :                               t ->               string = "cx_hstring_of"
  external ostring_of  :                               t ->               string = "cx_ostring_of"
  external nrandom     :                             int->                t      = "cx_f_nrandom"
  external zrandom     :                             int->                t      = "cx_f_zrandom"
  external nrandom1    :                             int->                t      = "cx_f_nrandom1"
  external zrandom1    :                             int->                t      = "cx_f_zrandom1"
  external nrandom_in  :               tref ->       int->                unit   = "cx_nrandom"
  external zrandom_in  :               tref ->       int->                unit   = "cx_zrandom"
  external nrandom1_in :               tref ->       int->                unit   = "cx_nrandom1"
  external zrandom1_in :               tref ->       int->                unit   = "cx_zrandom1"
  external xrandom_init:                             int->                unit   = "cx_random_init"
  external nbits       :                               t ->               int    = "cx_nbits"
  external lowbits     :                               t ->               int    = "cx_lowbits"
  external highbits    :                               t ->               int    = "cx_highbits"
  external nth_word    :                               t -> int ->        int    = "cx_nth_word"
  external nth_bit     :                               t -> int ->        bool   = "cx_nth_bit"
  external shl         :                               t -> int ->        t      = "cx_f_shl"
  external shr         :                               t -> int ->        t      = "cx_f_shr"
  external split       :                               t -> int ->        t*t    = "cx_f_split"
  external join        :                               t -> t   -> int -> t      = "cx_f_join"
  external shl_in      :               tref ->         t -> int ->        unit   = "cx_shl"
  external shr_in      :               tref ->         t -> int ->        unit   = "cx_shr"
  external split_in    :               tref -> tref -> t -> int ->        unit   = "cx_split"
  external join_in     :               tref ->         t -> t   -> int -> unit   = "cx_join"
  external dump        :                               t ->               unit   = "cx_dump"

  let zero   = of_int 0
  let one    = of_int 1

  let random_init x =
    xrandom_init x;
    if x = 0 then Random.self_init() else Random.init(x)

  (* affichage tronqu  1000 caractres *)
  let toplevel_print(a) =
    let s = string_of a      in
    let l = String.length(s) in
    if l < 1000 then Format.print_string s
    else begin
      Format.print_string (String.sub s 0 100);
      Format.print_string " ... (";
      Format.print_int    (l-200);
      Format.print_string " digits) ... ";
      Format.print_string (String.sub s (l-100) 100)
    end

  let toplevel_print_tref(a) =
    Format.print_string "tref(";
    toplevel_print(look a);
    Format.print_char ')'

  exception Error of string
  let _ = Callback.register_exception "cx_error" (Error "")

  external register : unit -> unit = "cx_register"
  let _ = register()

  external isprime     :                               t ->           tristate   = "cx_isprime"
  external isprime_1   :                             int ->           tristate   = "cx_isprime_1"
end

                           (* +-----------------+
                              |  module Slong   |
                              +-----------------+ *)

module Slong = struct

  type t
  type tref

  let name() = String.copy "Slong"

  external make_ref    :                               t ->               tref   = "sx_make_ref"
  external copy_in     :               tref ->         t ->               unit   = "sx_copy"
  external copy_out    :               tref ->                            t      = "sx_copy_out"
  external look        :               tref ->                            t      = "sx_look"
  external add         :                               t -> t   ->        t      = "sx_f_add"
  external add_1       :                               t -> int ->        t      = "sx_f_add_1"
  external add_in      :               tref ->         t -> t   ->        unit   = "sx_add"
  external add_1_in    :               tref ->         t -> int ->        unit   = "sx_add_1"
  external sub         :                               t -> t   ->        t      = "sx_f_sub"
  external sub_1       :                               t -> int ->        t      = "sx_f_sub_1"
  external sub_in      :               tref ->         t -> t   ->        unit   = "sx_sub"
  external sub_1_in    :               tref ->         t -> int ->        unit   = "sx_sub_1"
  external mul         :                               t -> t   ->        t      = "sx_f_mul"
  external mul_1       :                               t -> int ->        t      = "sx_f_mul_1"
  external mul_in      :               tref ->         t -> t   ->        unit   = "sx_mul"
  external mul_1_in    :               tref ->         t -> int ->        unit   = "sx_mul_1"
  external quomod      :                               t -> t   ->        t*t    = "sx_f_quomod"
  external quo         :                               t -> t   ->        t      = "sx_f_quo"
  external modulo      :                               t -> t   ->        t      = "sx_f_mod"
  external gquomod     : round_mode ->                 t -> t   ->        t*t    = "sx_f_gquomod"
  external gquo        : round_mode ->                 t -> t   ->        t      = "sx_f_gquo"
  external gmod        : round_mode ->                 t -> t   ->        t      = "sx_f_gmod"
  external quomod_in   :               tref -> tref -> t -> t   ->        unit   = "sx_quomod"
  external quo_in      :               tref ->         t -> t   ->        unit   = "sx_quo"
  external mod_in      :                       tref -> t -> t   ->        unit   = "sx_mod"
  external gquomod_in  : round_mode -> tref -> tref -> t -> t   ->        unit   = "sx_gquomod"
  external gquo_in     : round_mode -> tref ->         t -> t   ->        unit   = "sx_gquo"
  external gmod_in     : round_mode ->         tref -> t -> t   ->        unit   = "sx_gmod"
  external quomod_1    :                               t -> int ->        t*int  = "sx_f_quomod_1"
  external quo_1       :                               t -> int ->        t      = "sx_f_quo_1"
  external mod_1       :                               t -> int ->        int    = "sx_f_mod_1"
  external gquomod_1   : round_mode ->                 t -> int ->        t*int  = "sx_f_gquomod_1"
  external gquo_1      : round_mode ->                 t -> int ->        t      = "sx_f_gquo_1"
  external gmod_1      : round_mode ->                 t -> int ->        int    = "sx_f_gmod_1"
  external quomod_1_in :               tref ->         t -> int ->        int    = "sx_quomod_1"
  external quo_1_in    :               tref ->         t -> int ->        unit   = "sx_quo_1"
  external gquomod_1_in: round_mode -> tref ->         t -> int ->        int    = "sx_gquomod_1"
  external gquo_1_in   : round_mode -> tref ->         t -> int ->        unit   = "sx_gquo_1"
  external abs         :                               t ->               t      = "sx_f_abs"
  external abs_in      :               tref ->         t ->               unit   = "sx_abs"
  external neg         :                               t ->               t      = "sx_f_neg"
  external neg_in      :               tref ->         t ->               unit   = "sx_neg"
  external sqr         :                               t ->               t      = "sx_f_sqr"
  external pow         :                               t -> int ->        t      = "sx_f_pow"
  external pow_1       :                             int -> int ->        t      = "sx_f_pow_1"
  external powmod      :                               t -> t   -> t ->   t      = "sx_f_powmod"
  external gpowmod     : round_mode ->                 t -> t   -> t ->   t      = "sx_f_gpowmod"
  external sqr_in      :               tref ->         t ->               unit   = "sx_sqr"
  external pow_in      :               tref ->         t -> int ->        unit   = "sx_pow"
  external pow_1_in    :               tref ->       int -> int ->        unit   = "sx_pow_1"
  external powmod_in   :               tref ->         t -> t   -> t ->   unit   = "sx_powmod"
  external gpowmod_in  : round_mode -> tref ->         t -> t   -> t ->   unit   = "sx_gpowmod"
  external sqrt        :                               t ->               t      = "sx_f_sqrt"
  external root        :                               t -> int ->        t      = "sx_f_root"
  external gsqrt       : round_mode ->                 t ->               t      = "sx_f_gsqrt"
  external groot       : round_mode ->                 t -> int ->        t      = "sx_f_groot"
  external sqrt_in     :               tref ->         t ->               unit   = "sx_sqrt"
  external root_in     :               tref ->         t -> int ->        unit   = "sx_root"
  external gsqrt_in    : round_mode -> tref ->         t ->               unit   = "sx_gsqrt"
  external groot_in    : round_mode -> tref ->         t -> int ->        unit   = "sx_groot"
  external fact        :                             int ->               t      = "sx_f_fact"
  external fact_in     :               tref ->       int ->               unit   = "sx_fact"
  external gcd         :                               t -> t   ->        t      = "sx_f_gcd"
  external gcd_ex      :                               t -> t   ->        t*t*t  = "sx_f_gcd_ex"
  external cfrac       :                               t -> t   ->     t*t*t*t*t = "sx_f_cfrac"
  external gcd_in      : tref->                        t -> t   ->        unit   = "sx_gcd"
  external gcd_ex_in   : tref->tref->tref->            t -> t   ->        unit   = "sx_gcd_ex"
  external cfrac_in    : tref->tref->tref->tref->tref->t -> t   ->        unit   = "sx_cfrac_bytecode" "sx_cfrac"
  external isprime     :                               t ->           tristate   = "sx_isprime"
  external isprime_1   :                             int ->           tristate   = "sx_isprime_1"
  external sgn         :                               t ->               int    = "sx_sgn"
  external cmp         :                               t -> t   ->        int    = "sx_cmp"
  external cmp_1       :                               t -> int ->        int    = "sx_cmp_1"
  external eq          :                               t -> t   ->        bool   = "sx_eq"
  external eq_1        :                               t -> int ->        bool   = "sx_eq_1"
  external neq         :                               t -> t   ->        bool   = "sx_neq"
  external neq_1       :                               t -> int ->        bool   = "sx_neq_1"
  external inf         :                               t -> t   ->        bool   = "sx_inf"
  external inf_1       :                               t -> int ->        bool   = "sx_inf_1"
  external infeq       :                               t -> t   ->        bool   = "sx_infeq"
  external infeq_1     :                               t -> int ->        bool   = "sx_infeq_1"
  external sup         :                               t -> t   ->        bool   = "sx_sup"
  external sup_1       :                               t -> int ->        bool   = "sx_sup_1"
  external supeq       :                               t -> t   ->        bool   = "sx_supeq"
  external supeq_1     :                               t -> int ->        bool   = "sx_supeq_1"
  external of_int      :                             int ->               t      = "sx_of_int"
  external of_string   :                          string ->               t      = "sx_of_string"
  external of_int_in   :               tref ->       int ->               unit   = "sx_copy_int"
  external of_string_in:               tref ->    string ->               unit   = "sx_copy_string"
  external int_of      :                               t ->               int    = "sx_int_of"
  external string_of   :                               t ->               string = "sx_string_of"
  external bstring_of  :                               t ->               string = "sx_bstring_of"
  external hstring_of  :                               t ->               string = "sx_hstring_of"
  external ostring_of  :                               t ->               string = "sx_ostring_of"
  external nrandom     :                             int->                t      = "sx_f_nrandom"
  external zrandom     :                             int->                t      = "sx_f_zrandom"
  external nrandom1    :                             int->                t      = "sx_f_nrandom1"
  external zrandom1    :                             int->                t      = "sx_f_zrandom1"
  external nrandom_in  :               tref ->       int->                unit   = "sx_nrandom"
  external zrandom_in  :               tref ->       int->                unit   = "sx_zrandom"
  external nrandom1_in :               tref ->       int->                unit   = "sx_nrandom1"
  external zrandom1_in :               tref ->       int->                unit   = "sx_zrandom1"
  external xrandom_init:                             int->                unit   = "sx_random_init"
  external nbits       :                               t ->               int    = "sx_nbits"
  external lowbits     :                               t ->               int    = "sx_lowbits"
  external highbits    :                               t ->               int    = "sx_highbits"
  external nth_word    :                               t -> int ->        int    = "sx_nth_word"
  external nth_bit     :                               t -> int ->        bool   = "sx_nth_bit"
  external shl         :                               t -> int ->        t      = "sx_f_shl"
  external shr         :                               t -> int ->        t      = "sx_f_shr"
  external split       :                               t -> int ->        t*t    = "sx_f_split"
  external join        :                               t -> t   -> int -> t      = "sx_f_join"
  external shl_in      :               tref ->         t -> int ->        unit   = "sx_shl"
  external shr_in      :               tref ->         t -> int ->        unit   = "sx_shr"
  external split_in    :               tref -> tref -> t -> int ->        unit   = "sx_split"
  external join_in     :               tref ->         t -> t   -> int -> unit   = "sx_join"
  external dump        :                               t ->               unit   = "sx_dump"

  let zero   = of_int 0
  let one    = of_int 1

  let random_init x =
    xrandom_init x;
    if x = 0 then Random.self_init() else Random.init(x)

  (* affichage tronqu  1000 caractres *)
  let toplevel_print(a) =
    let s = string_of a      in
    let l = String.length(s) in
    if l < 1000 then Format.print_string s
    else begin
      Format.print_string (String.sub s 0 100);
      Format.print_string " ... (";
      Format.print_int    (l-200);
      Format.print_string " digits) ... ";
      Format.print_string (String.sub s (l-100) 100)
    end

  let toplevel_print_tref(a) =
    Format.print_string "tref(";
    toplevel_print(look a);
    Format.print_char ')'

  exception Error of string
  let _ = Callback.register_exception "sx_error" (Error "")

  external register : unit -> unit = "sx_register"
  let _ = register()

  external isprime     :                               t ->           tristate   = "sx_isprime"
  external isprime_1   :                             int ->           tristate   = "sx_isprime_1"
end

                           (* +----------------+
                              |   module Big   |
                              +----------------+ *)

module Big = struct
  open Big_int

  type t     = big_int
  type tref  = t ref
  let name() = String.copy "Big"
  let zero   = zero_big_int
  let one    = unit_big_int
  let two    = big_int_of_int 2
  exception Error of string

  (* gestion des rfrences *)
  let make_ref   a = ref a
  let copy_in  r a = r := a
  let copy_out r   = !r
  let look     r   = !r

  (* addition/soustraction/multiplication *)
  let abs        a   =      abs_big_int      a
  let neg        a   =      minus_big_int    a
  let add        a b =      add_big_int      a b
  let add_1      a b =      add_int_big_int  b a
  let sub        a b =      sub_big_int      a b
  let sub_1      a b =      sub_big_int      a (big_int_of_int b)
  let mul        a b =      mult_big_int     a b
  let mul_1      a b =      mult_int_big_int b a
  let sqr        a   =      square_big_int   a
  let abs_in   r a   = r := abs_big_int      a
  let neg_in   r a   = r := minus_big_int    a
  let add_in   r a b = r := add_big_int      a b
  let add_1_in r a b = r := add_int_big_int  b a
  let sub_in   r a b = r := sub_big_int      a b
  let sub_1_in r a b = r := sub_big_int      a (big_int_of_int b)
  let mul_in   r a b = r := mult_big_int     a b
  let mul_1_in r a b = r := mult_int_big_int b a
  let sqr_in   r a   = r := square_big_int   a

  (* division euclidienne *)
  let quomod     a b = if sign_big_int(b) > 0
                       then quomod_big_int a b
                       else let a1 = minus_big_int a
                            and b1 = minus_big_int b in
                            let q,r = quomod_big_int a1 b1
                            in (q,minus_big_int r)

  let quo        a b = let q,_ = quomod a b in q
  let modulo     a b = let _,r = quomod a b in r

  let gquomod m  a b = match m with

    | Floor        -> quomod a b

    | Ceil         -> let q,r = quomod a b in if sign_big_int(r) = 0
                      then q,r
                      else (succ_big_int q),(sub_big_int r b)

    | Nearest_up   -> let a1 = add_big_int a (add_big_int a b)
                      and b1 = add_big_int b b in
                      let q,r = quomod a1 b1 in
                      (q, div_big_int (sub_big_int r b) two)

    | Nearest_down -> let a1 = add_big_int a (add_big_int a b)
                      and b1 = add_big_int b b in
                      let q,r = quomod a1 b1 in if sign_big_int(r) = 0
                      then (pred_big_int q, div_big_int (minus_big_int b) two)
                      else (q, div_big_int (sub_big_int r b) two)

  let gquo       m     a b = let q,_ = gquomod m a b in q
  let gmod       m     a b = let _,r = gquomod m a b in r
  let quomod_in    r s a b = let x,y = quomod    a b in r := x; s := y
  let quo_in       r   a b = let x,_ = quomod    a b in r := x
  let mod_in         s a b = let _,y = quomod    a b in s := y
  let gquomod_in m r s a b = let x,y = gquomod m a b in r := x; s := y
  let gquo_in    m r   a b = let x,_ = gquomod m a b in r := x
  let gmod_in    m   s a b = let _,y = gquomod m a b in s := y

  let quomod_1         a b = let q,r = quomod    a (big_int_of_int b) in q,     (int_of_big_int r)
  let quo_1            a b = let q,_ = quomod    a (big_int_of_int b) in q
  let mod_1            a b = let _,r = quomod    a (big_int_of_int b) in        (int_of_big_int r)
  let gquomod_1      m a b = let q,r = gquomod m a (big_int_of_int b) in q,     (int_of_big_int r)
  let gquo_1         m a b = let q,_ = gquomod m a (big_int_of_int b) in q
  let gmod_1         m a b = let _,r = gquomod m a (big_int_of_int b) in         int_of_big_int r
  let quomod_1_in    r a b = let x,y = quomod    a (big_int_of_int b) in r := x; int_of_big_int y
  let quo_1_in       r a b = let x,y = quomod    a (big_int_of_int b) in r := x
  let gquomod_1_in m r a b = let x,y = gquomod m a (big_int_of_int b) in r := x; int_of_big_int y
  let gquo_1_in    m r a b = let x,y = gquomod m a (big_int_of_int b) in r := x

  (* reprsentation binaire *)
  let words_in_digit = Nat.length_of_digit/16
  let nat_of b = (Obj.obj(Obj.field (Obj.repr b) 1) : Nat.nat)

  let nth_word x n =
    if n < 0 then 0
    else let n1 = n/words_in_digit
         and n2 = 16*(n mod words_in_digit) in
         if n1 >= num_digits_big_int x then 0
         else begin
           let tmp = Nat.create_nat 2 in
           Nat.blit_nat tmp 1 (nat_of x) n1 1;
           Nat.shift_right_nat tmp 1 1 tmp 0 n2;
           (Nat.nth_digit_nat tmp 1) land 0xffff
         end

  let masque = [| 0x0001; 0x0002; 0x0004; 0x0008;
                  0x0010; 0x0020; 0x0040; 0x0080;
                  0x0100; 0x0200; 0x0400; 0x0800;
                  0x1000; 0x2000; 0x4000; 0x8000 |]
  let nth_bit n k =
    (k >= 0) && ((nth_word n (k/16)) land masque.(k land 15) <> 0)

 let nbits x =
    let a = nat_of x             in
    let l = num_digits_big_int x in
    l * Nat.length_of_digit - (Nat.num_leading_zero_bits_in_digit a (l-1))

  let lowbits x = (Nat.nth_digit_nat (nat_of x) 0) land 0x7fffffff

  let highbits x =
    let a = nat_of x
    and l = num_digits_big_int x in
    let m = Nat.length_of_digit - (Nat.num_leading_zero_bits_in_digit a (l-1)) in
    let tmp = Nat.create_nat 2 in
    if l < 2 then begin
      Nat.blit_nat tmp 1 a (l-1) 1;
      Nat.set_digit_nat tmp 0 0;
    end
    else Nat.blit_nat tmp 0 a (l-2) 2;
    if m < 31 then Nat.shift_left_nat  tmp 0 2 tmp 0 (31-m)
              else Nat.shift_right_nat tmp 1 1 tmp 0 (m-31);
    Nat.nth_digit_nat tmp 1

  (* dcalages *)
  let shl a b =
    if (sign_big_int(a) = 0) or (b <= -nbits(a)) then zero_big_int
    else if b = 0 then a
    else if b > 0 then mult_big_int a (power_int_positive_int 2 b)
    else if sign_big_int(a) > 0 then div_big_int a (power_int_positive_int 2 (-b))
    else minus_big_int(div_big_int (minus_big_int a) (power_int_positive_int 2 (-b)))

  let shr a b =
    if (sign_big_int(a) = 0) or (b >= nbits(a)) then zero_big_int
    else if b = 0 then a
    else if b < 0 then mult_big_int a (power_int_positive_int 2 (-b))
    else if sign_big_int(a) > 0 then div_big_int a (power_int_positive_int 2 b)
    else minus_big_int(div_big_int (minus_big_int a) (power_int_positive_int 2 b))

  let split a b =
    if b < 0 then raise (Error "negative index")
    else if (b = 0) or (sign_big_int(a) = 0) then a,zero_big_int
    else if b >= nbits(a)                    then zero_big_int,a
    else let q,r = quomod_big_int (abs_big_int a) (power_int_positive_int 2 b) in
         if sign_big_int(a) > 0 then q,r else minus_big_int(q),minus_big_int(r)

  let join a b c =
    if c < 0 then raise (Error "negative index") else add a (shl b c)

  let shl_in   r   a b = r := shl a b
  let shr_in   r   a b = r := shr a b
  let split_in r s a b = let x,y = split a b in r := x; s := y
  let join_in  r a b c = r := join a b c

  (* exponentiation *)

  (*let pow        a b =      power_big_int_positive_int a b *)
  (* bogu : 416577169024505710327332685500 ^ 45 -> ...442939392 *)

  let pow a b =
    if b < 0 then raise (Error "negative exponent");
    let rec loop n =
      if n = 0 then unit_big_int
      else let x = square_big_int(loop (n/2)) in
           if n land 1 = 0 then x else mult_big_int a x
    in loop b

  let pow_in   r a b = r := pow a b
  let pow_1      a b =      power_int_positive_int     a b
  let pow_1_in r a b = r := power_int_positive_int     a b

  let gpowmod m a b c =
    if sign_big_int(b) < 0 then raise (Error "negative exponent");
    let r = ref(unit_big_int) in
    for i=nbits(b)-1 downto 0 do
      r := mod_big_int (square_big_int !r) c;
      if nth_bit b i then r := mod_big_int (mult_big_int !r a) c;
    done;
    gmod m !r c

  let powmod         a b c =      gpowmod Floor a b c
  let powmod_in    r a b c = r := gpowmod Floor a b c
  let gpowmod_in m r a b c = r := gpowmod m     a b c
        
  (* racine carre *)
  let sqrt    a = sqrt_big_int a
  let gsqrt m a = match m with

    | Floor -> sqrt_big_int a

    | Ceil  -> let b = sqrt_big_int a in
               let c = sub_big_int a (square_big_int b) in
               if sign_big_int(c) = 0 then b else succ_big_int(b)

    | _     -> let a1 = mult_int_big_int 4 a in
               let b1 = sqrt_big_int a1 in
               let b,r = quomod_big_int b1 two in
               if sign_big_int(r) = 0 then b else succ_big_int b

  let sqrt_in    r a = r := sqrt_big_int a
  let gsqrt_in m r a = r := gsqrt m a

  (* racine p-me *)
  let root a p =

    let p1 = p-1 in

    (* itration de Newton : cas a > 0 *)
    let rec loop_positif a x =
      let y = pow x p1  in
      let z = sub_big_int a (mult_big_int x y) in
      if sign_big_int(z) >= 0 then x
      else loop_positif a (add_big_int x (div_big_int z (mult_int_big_int p y)))

    (* itration de Newton : cas a < 0 *)
    and loop_negatif a x =
      let y = pow x p1  in
      let z = sub_big_int a (mult_big_int x y) in
      match sign_big_int(z) with
        | 0  -> x
        | -1 -> pred_big_int x
        | _  -> let y = mult_int_big_int p y
                in loop_negatif a (add_big_int x (quo (add_big_int z (pred_big_int y)) y))
    in

    (* racine rcursive *)
    let rec root_rec a =

      (* valeur initiale : puissance de 2 si a est petit, *)
      (* sinon racine p-me de la partie haute            *)
      let init =
        let n = nbits(a)  in
        let q = n/(2*p)   in
        if q < 2 then shl (big_int_of_int (sign_big_int a)) ((n+p1)/p)
                 else shl (add_int_big_int (sign_big_int a) (root_rec (shr a (p*q)))) q
      in
      (if sign_big_int(a) >= 0 then loop_positif else loop_negatif) a init
    in

    (* contrle de signe *)
    if p <= 0                                   then raise (Error "negative exponent");
    if ((p mod 2) = 0) && (sign_big_int(a) < 0) then raise (Error "negative base");
    if (sign_big_int(a) = 0) then zero_big_int else root_rec a

  let groot m a p = match m with
    | Floor -> root a p
    | Ceil  -> let r = root a p in
               let b = sub_big_int a (pow r p) in
                 if sign_big_int(b) = 0 then r else succ_big_int r
    | _     -> let r = root (shl a p) p in div_big_int (succ_big_int r) two

  let root_in    r a p = r := root    a p
  let groot_in m r a p = r := groot m a p

  (* factorielle *)
  let fact a = 
    if a < 0 then raise (Error "negative base")
    else begin
      let r = ref(unit_big_int) in
      for i=2 to a do r := mul !r (big_int_of_int i) done;
      !r
    end
  let fact_in r a = r := fact a

  (* pgcd *)
  let gcd a b = gcd_big_int a b

  let rec gcd_ex a b = match sign_big_int(a) with

    | 0 -> if sign_big_int(b) >= 0
              then b,zero_big_int,big_int_of_int(-1)
              else minus_big_int(b),zero_big_int,unit_big_int

    | 1 -> let q,r = quomod b a      in
           let d,u1,v1 = gcd_ex r a  in
           let u = minus_big_int(u1) in d,(sub_big_int (mult_big_int q u) v1),u

    |_  -> let d,u,v = gcd_ex (minus_big_int a) b in d,(minus_big_int u),v

  let cfrac a b =
    if (sign_big_int(a) = 0) & (sign_big_int(b) = 0)
    then zero_big_int,unit_big_int,zero_big_int,unit_big_int,zero_big_int
    else let d,u,v = gcd_ex a b in d,u,v,(div_big_int a d),(div_big_int b d)

  let gcd_in    d         a b = d := gcd_big_int a b
  let gcd_ex_in d u v     a b = let     x,y,z = gcd_ex a b in d := x; u := y; v := z
  let cfrac_in  d u v p q a b = let r,s,x,y,z = cfrac  a b in d := r; u := s; v := x; p := y; q := z

  (* comparaisons *)
  let sgn     a   = sign_big_int a
  let cmp     a b = compare_big_int a b
  let eq      a b = eq_big_int      a b
  let neq     a b = not(eq_big_int  a b)
  let inf     a b = lt_big_int      a b
  let infeq   a b = le_big_int      a b
  let sup     a b = gt_big_int      a b
  let supeq   a b = ge_big_int      a b
  let cmp_1   a b = compare_big_int a (big_int_of_int b)
  let eq_1    a b = eq_big_int      a (big_int_of_int b)
  let neq_1   a b = not(eq_big_int  a (big_int_of_int b))
  let inf_1   a b = lt_big_int      a (big_int_of_int b)
  let infeq_1 a b = le_big_int      a (big_int_of_int b)
  let sup_1   a b = gt_big_int      a (big_int_of_int b)
  let supeq_1 a b = ge_big_int      a (big_int_of_int b)

  (* conversions *)
  let of_int      a =      big_int_of_int a
  let of_int_in r a = r := big_int_of_int a

  let of_string ch =

    let l = String.length(ch)
    and i = ref(0) in

    (* dcode le signe *)
    let s =      if (l > 0) & ch.[0] = '-' then (incr i; -1)
            else if (l > 0) & ch.[0] = '+' then (incr i;  1) else 1 in

    (* dcode la base *)
    let b = if (l > !i+1) & (ch.[!i] = '0') then match ch.[!i+1] with
            | 'x' | 'X' -> i := !i + 2; 16
            | 'o' | 'O' -> i := !i + 2;  8
            | 'b' | 'B' -> i := !i + 2;  2
            | _         -> incr i;      10
            else 10 in

    (* s'il n'y a plus de caractres disponibles, chane invalide *)
    if !i >= l then raise (Error "invalid string");

    (* dcode le reste de la chane (algorithme en n^2) *)
    let r = ref(zero_big_int) in
    while !i < l do
      let c = match ch.[!i] with
      | '0'..'9' as x -> Char.code(x)-48
      | 'a'..'f' as x -> Char.code(x)-87
      | 'A'..'F' as x -> Char.code(x)-55
      | _             -> raise (Error "invalid string")
      in
      if c > b then raise (Error "invalid string");
      r := add_int_big_int c (mult_int_big_int b !r);
      incr i
    done;

    (* rsultat *)
    if s = 1 then !r else minus_big_int !r

  let of_string_in r ch = r := of_string ch

  let int_of a    = int_of_big_int a
  let string_of a = string_of_big_int a

  let bstring_of a =
    if sign_big_int(a) = 0 then "0"
    else begin
      let n = nbits(a)
      and s = if sign_big_int(a) < 0 then 1 else 0 in
      let r = String.create (n+2+s) in
        if (s=1) then r.[0] <- '-';
        r.[s]   <- '0';
        r.[s+1] <- 'b';
        for i=0 to n-1 do r.[n+1+s-i] <- if nth_bit a i then '1' else '0' done;
        r
    end

  let hstring_of a =
    if sign_big_int(a) = 0 then "0"
    else begin
      let n = (nbits(a)+3)/4
      and s = if sign_big_int(a) < 0 then 1 else 0 in
      let r = String.create (n+2+s) in
        if (s=1) then r.[0] <- '-';
        r.[s]   <- '0';
        r.[s+1] <- 'x';
        for i=0 to n-1 do
          let c = ((nth_word a (i/4)) lsr (4*(i land 3))) land 15 in
          r.[n+1+s-i] <- if c < 10 then Char.chr(48+c) else Char.chr(55+c)
        done;
        r
    end

  let ostring_of a =
    if sign_big_int(a) = 0 then "0"
    else begin
      let n = (nbits(a)+2)/3
      and s = if sign_big_int(a) < 0 then 1 else 0 in
      let r = String.create (n+2+s) in
        if (s=1) then r.[0] <- '-';
        r.[s]   <- '0';
        r.[s+1] <- 'o';
        for i=0 to n-1 do
          let c = (if nth_bit a (3*i)   then 1 else 0)
                + (if nth_bit a (3*i+1) then 1 else 0)*2
                + (if nth_bit a (3*i+2) then 1 else 0)*4 in
          r.[n+1+s-i] <- Char.chr(48+c)
        done;
        r
    end

  (* nombres alatoires *)
  let nrandom n =
    if n < 0 then raise (Error "negative size")
    else if n = 0 then zero_big_int
    else begin
      let p = if n land 15 = 0 then 16 else n land 15 in
      let mask = (1 lsl p) - 1 in
      let r = ref(big_int_of_int(Random.bits() land mask)) in
      for i=1 to (n-p)/16 do
        r := add_int_big_int (Random.bits() land 0xffff) (mult_int_big_int 0x10000 !r)
      done;
      !r
    end

  let nrandom1 n =
    if n < 0 then raise (Error "negative size")
    else if n = 0 then zero_big_int
    else begin
      let p = if n land 15 = 0 then 16 else n land 15 in
      let mask = 1 lsl (p-1) in
      let r = ref(big_int_of_int((Random.bits() land (mask-1)) lor mask)) in
      for i=1 to (n-p)/16 do
        r := add_int_big_int (Random.bits() land 0xffff) (mult_int_big_int 0x10000 !r)
      done;
      !r
    end

  let zrandom n = match Random.bits() land 1 with
    | 0 -> nrandom n
    | _ -> minus_big_int(nrandom n)

  let zrandom1 n = match Random.bits() land 1 with
    | 0 -> nrandom1 n
    | _ -> minus_big_int(nrandom1 n)

  let nrandom_in  r n = r := nrandom  n
  let zrandom_in  r n = r := zrandom  n
  let nrandom1_in r n = r := nrandom1 n
  let zrandom1_in r n = r := zrandom1 n

  let random_init x = if x = 0 then Random.self_init() else Random.init(x)

  (* affichage tronqu  1000 caractres *)
  let toplevel_print(a) =
    let s = string_of a      in
    let l = String.length(s) in
    if l < 1000 then Format.print_string s
    else begin
      Format.print_string (String.sub s 0 100);
      Format.print_string " ... (";
      Format.print_int    (l-200);
      Format.print_string " digits) ... ";
      Format.print_string (String.sub s (l-100) 100)
    end

  let toplevel_print_tref(a) =
    Format.print_string "tref(";
    toplevel_print(look a);
    Format.print_char ')'

                         (* +---------------------+
                            |  Test de primalit  |
                            +---------------------+ *)

  (* n >= 0, cherche s'il a un diviseur non trivial <= 2^10 *)
  let trivial_prime n =

    let rec loop d =
      if inf_1 n (d*d)      then True
      else if d > 1024      then Unknown
      else if mod_1 n d = 0 then False
      else loop(d+2)
    in

    match cmp_1 n 2 with
    |  1 when nth_bit n 0 -> loop 3
    |  0 -> True
    |  _ -> False

  (* test de Rabin-Miller pour les bases 2,3,5,7,11,13,17.
     on suppose n >= 0 et trivial_prime(n) = Unknown
     (ie. n >= 2^20 et n n'a pas de facteur premier <= 2^10)

     Ribenboim, The Little Book of Bigger Primes, ch. VII p. 98 :

     Given k >= 1, denote by tk the smallest integer which is a strong pseudoprime
     for the bases p1 = 2, p2 = 3, . . . , pk, simultaneously. Then the
     calculations of Pomerance, Selfridge & Wagstaff (1980), extended by
     Jaeschke (1993), provide the following values:

     t2 =                1 373 653 = 829  1657,
     t3 =               25 326 001 = 2251  11251,
     t4 =            3 215 031 751 = 151  751  28351,
     t5 =        2 152 302 898 747 = 6763  10627  29947,
     t6 =        3 474 749 660 383 = 1303  16927  157543,
     t7 = t8 = 341 550 071 728 321 = 10670053  32010157.

     Complment (http://primes.utm.edu/prove/prove2_3.html) : le premier
     contre-exemple suivant t4 est

     t4' = 118 670 087 467 = 172243  688969.

     On utilisera cette borne car t4 est reconnu non premier par trivial_prime.
     Le mme argument s'applique  t2, avec

     t2' = 2 284 453 = 1069  2137.

     Le premier nombre sans facteur trivial spp(2) est

     t1' = 1 194 649 = 1093^2.

     Quelques contre-exemples (thse de Franois Arnault,
     http://www.unilim.fr/pages_perso/francois.arnault/documents/these.ps):

     1195068768795265792518361315725116351898245581
     = 24444516448431392447461  48889032896862784894921
     est spp(2,..,31)

     80383745745363949125707961434194210813883768828755\
     81458374889175222974273765333652186502336163960045\
     45791504202360320876656996676098728404396540823292\
     87387918508691668573282677617710293896977394701670\
     82304286871099974399765441448453411558724506334092\
     79022275296229414984230688168540432645753401832978\
     6111298960644845216191652872597534901

     = 20047910831974980270915322604227342650259408302\
     05662543872531023690016085350598121358111595798609\
     86679108158254267908348457261690695858464376399022\
     2898400226296015918301

      400958216639499605418306452084546853005188166041\
     13250877450620473800321707011962427162231915972197\
     33582163165085358166969145233813917169287527980445\
     796800452592031836601

     est spp(2,..,199)

     Ces deux nombres sont reconnus non premiers par small_prime en comparant les
     racines carres de -1 obtenues aux cours des tests. De mme, tous les spp
     donns dans la thse de Franois Arnault page 29 sont reconnus non premiers
     par le mme critre  l'exception de

     n = 340770619658554299579514119954792724101734599955221
     = 13053172404794060245721341  26106344809588120491442681

     qui est spp(2,..,31). On aurait trouv une deuxime racine carre
     de -1 si on avait pouss le test jusqu' la base 23.
  *)

  let miller_rabin =

    let tests = [of_int  2, of_string         "1194649";
                 of_int  3, of_string         "2284453";
		 of_int  5, of_string        "25326001";
		 of_int  7, of_string    "118670087467";
		 of_int 11, of_string   "2152302898747";
		 of_int 13, of_string   "3474749660383";
		 of_int 17, of_string "341550071728321"] in

    fun n ->

      (* dcompose n-1 = q*2^k, q impair *)
      let q,k =
	let rec loop k = if nth_bit n k then (shr n k),k else loop (k+1)
	in loop 1
      in

      (* racine de -1 potentielle *)
      let i = make_ref(zero) in

      (* lve a au carr au plus k-2 fois jusqu' trouver une racine carre de -1 *)
      let rec sqr_loop a k =
	if k <= 1 then None
	else let b = gmod Nearest_up (sqr a) n in
	     if eq_1 b (-1) then Some(if sgn a > 0 then a else neg a)
	     else sqr_loop b (k-1)
      in

      (* test de Rabin-Miller pour la base a *)
      let mr_test a =
	let b = gpowmod Nearest_up a q n in
	(eq_1 b 1) or (eq_1 b (-1)) or match sqr_loop b k with
	| Some(x) when eq_1 (look i) 0 -> copy_in i x; true
	| Some(x) -> eq (look i) x
	| None    -> false
      in

      (* passe en revue toutes les bases jusqu' obtenir une certitude *)
      let rec loop l = match l with
      | [] -> Unknown
      | (a,t)::suite -> match cmp t n with
	  | 0 -> False
	  | 1 -> if mr_test a then True else False
	  | _ -> if mr_test a then loop suite else False
      in loop tests

  (* test de primalit *)
  let isprime n =
    let n = if inf_1 n 0 then neg n else n in
    match trivial_prime n with Unknown -> miller_rabin n | r -> r

  let isprime_1 n = isprime(of_int n)

end

                           (* +-----------------+
                              |  module Gmp   |
                              +-----------------+ *)

module Gmp = struct

  type t
  type tref

  let name() = String.copy "Gmp"

  external make_ref    :                               t ->               tref   = "gx_make_ref"
  external copy_in     :               tref ->         t ->               unit   = "gx_copy"
  external copy_out    :               tref ->                            t      = "gx_copy_out"
  external look        :               tref ->                            t      = "gx_look"
  external add         :                               t -> t   ->        t      = "gx_f_add"
  external add_1       :                               t -> int ->        t      = "gx_f_add_1"
  external add_in      :               tref ->         t -> t   ->        unit   = "gx_add"
  external add_1_in    :               tref ->         t -> int ->        unit   = "gx_add_1"
  external sub         :                               t -> t   ->        t      = "gx_f_sub"
  external sub_1       :                               t -> int ->        t      = "gx_f_sub_1"
  external sub_in      :               tref ->         t -> t   ->        unit   = "gx_sub"
  external sub_1_in    :               tref ->         t -> int ->        unit   = "gx_sub_1"
  external mul         :                               t -> t   ->        t      = "gx_f_mul"
  external mul_1       :                               t -> int ->        t      = "gx_f_mul_1"
  external mul_in      :               tref ->         t -> t   ->        unit   = "gx_mul"
  external mul_1_in    :               tref ->         t -> int ->        unit   = "gx_mul_1"
  external quomod      :                               t -> t   ->        t*t    = "gx_f_quomod"
  external quo         :                               t -> t   ->        t      = "gx_f_quo"
  external modulo      :                               t -> t   ->        t      = "gx_f_mod"
  external gquomod     : round_mode ->                 t -> t   ->        t*t    = "gx_f_gquomod"
  external gquo        : round_mode ->                 t -> t   ->        t      = "gx_f_gquo"
  external gmod        : round_mode ->                 t -> t   ->        t      = "gx_f_gmod"
  external quomod_in   :               tref -> tref -> t -> t   ->        unit   = "gx_quomod"
  external quo_in      :               tref ->         t -> t   ->        unit   = "gx_quo"
  external mod_in      :                       tref -> t -> t   ->        unit   = "gx_mod"
  external gquomod_in  : round_mode -> tref -> tref -> t -> t   ->        unit   = "gx_gquomod"
  external gquo_in     : round_mode -> tref ->         t -> t   ->        unit   = "gx_gquo"
  external gmod_in     : round_mode ->         tref -> t -> t   ->        unit   = "gx_gmod"
  external quomod_1    :                               t -> int ->        t*int  = "gx_f_quomod_1"
  external quo_1       :                               t -> int ->        t      = "gx_f_quo_1"
  external mod_1       :                               t -> int ->        int    = "gx_f_mod_1"
  external gquomod_1   : round_mode ->                 t -> int ->        t*int  = "gx_f_gquomod_1"
  external gquo_1      : round_mode ->                 t -> int ->        t      = "gx_f_gquo_1"
  external gmod_1      : round_mode ->                 t -> int ->        int    = "gx_f_gmod_1"
  external quomod_1_in :               tref ->         t -> int ->        int    = "gx_quomod_1"
  external quo_1_in    :               tref ->         t -> int ->        unit   = "gx_quo_1"
  external gquomod_1_in: round_mode -> tref ->         t -> int ->        int    = "gx_gquomod_1"
  external gquo_1_in   : round_mode -> tref ->         t -> int ->        unit   = "gx_gquo_1"
  external abs         :                               t ->               t      = "gx_f_abs"
  external abs_in      :               tref ->         t ->               unit   = "gx_abs"
  external neg         :                               t ->               t      = "gx_f_neg"
  external neg_in      :               tref ->         t ->               unit   = "gx_neg"
  external sqr         :                               t ->               t      = "gx_f_sqr"
  external pow         :                               t -> int ->        t      = "gx_f_pow"
  external pow_1       :                             int -> int ->        t      = "gx_f_pow_1"
  external powmod      :                               t -> t   -> t ->   t      = "gx_f_powmod"
  external gpowmod     : round_mode ->                 t -> t   -> t ->   t      = "gx_f_gpowmod"
  external sqr_in      :               tref ->         t ->               unit   = "gx_sqr"
  external pow_in      :               tref ->         t -> int ->        unit   = "gx_pow"
  external pow_1_in    :               tref ->       int -> int ->        unit   = "gx_pow_1"
  external powmod_in   :               tref ->         t -> t   -> t ->   unit   = "gx_powmod"
  external gpowmod_in  : round_mode -> tref ->         t -> t   -> t ->   unit   = "gx_gpowmod"
  external sqrt        :                               t ->               t      = "gx_f_sqrt"
  external root        :                               t -> int ->        t      = "gx_f_root"
  external gsqrt       : round_mode ->                 t ->               t      = "gx_f_gsqrt"
  external groot       : round_mode ->                 t -> int ->        t      = "gx_f_groot"
  external sqrt_in     :               tref ->         t ->               unit   = "gx_sqrt"
  external root_in     :               tref ->         t -> int ->        unit   = "gx_root"
  external gsqrt_in    : round_mode -> tref ->         t ->               unit   = "gx_gsqrt"
  external groot_in    : round_mode -> tref ->         t -> int ->        unit   = "gx_groot"
  external fact        :                             int ->               t      = "gx_f_fact"
  external fact_in     :               tref ->       int ->               unit   = "gx_fact"
  external gcd         :                               t -> t   ->        t      = "gx_f_gcd"
  external gcd_ex      :                               t -> t   ->        t*t*t  = "gx_f_gcd_ex"
  external cfrac       :                               t -> t   ->     t*t*t*t*t = "gx_f_cfrac"
  external gcd_in      : tref->                        t -> t   ->        unit   = "gx_gcd"
  external gcd_ex_in   : tref->tref->tref->            t -> t   ->        unit   = "gx_gcd_ex"
  external cfrac_in    : tref->tref->tref->tref->tref->t -> t   ->        unit   = "gx_cfrac_bytecode" "gx_cfrac"
  external isprime     :                               t ->           tristate   = "gx_isprime"
  external isprime_1   :                             int ->           tristate   = "gx_isprime_1"
  external sgn         :                               t ->               int    = "gx_sgn"
  external cmp         :                               t -> t   ->        int    = "gx_cmp"
  external cmp_1       :                               t -> int ->        int    = "gx_cmp_1"
  external eq          :                               t -> t   ->        bool   = "gx_eq"
  external eq_1        :                               t -> int ->        bool   = "gx_eq_1"
  external neq         :                               t -> t   ->        bool   = "gx_neq"
  external neq_1       :                               t -> int ->        bool   = "gx_neq_1"
  external inf         :                               t -> t   ->        bool   = "gx_inf"
  external inf_1       :                               t -> int ->        bool   = "gx_inf_1"
  external infeq       :                               t -> t   ->        bool   = "gx_infeq"
  external infeq_1     :                               t -> int ->        bool   = "gx_infeq_1"
  external sup         :                               t -> t   ->        bool   = "gx_sup"
  external sup_1       :                               t -> int ->        bool   = "gx_sup_1"
  external supeq       :                               t -> t   ->        bool   = "gx_supeq"
  external supeq_1     :                               t -> int ->        bool   = "gx_supeq_1"
  external of_int      :                             int ->               t      = "gx_of_int"
  external of_string   :                          string ->               t      = "gx_of_string"
  external of_int_in   :               tref ->       int ->               unit   = "gx_copy_int"
  external of_string_in:               tref ->    string ->               unit   = "gx_copy_string"
  external int_of      :                               t ->               int    = "gx_int_of"
  external string_of   :                               t ->               string = "gx_string_of"
  external bstring_of  :                               t ->               string = "gx_bstring_of"
  external hstring_of  :                               t ->               string = "gx_hstring_of"
  external ostring_of  :                               t ->               string = "gx_ostring_of"
  external nrandom     :                             int->                t      = "gx_f_nrandom"
  external zrandom     :                             int->                t      = "gx_f_zrandom"
  external nrandom1    :                             int->                t      = "gx_f_nrandom1"
  external zrandom1    :                             int->                t      = "gx_f_zrandom1"
  external nrandom_in  :               tref ->       int->                unit   = "gx_nrandom"
  external zrandom_in  :               tref ->       int->                unit   = "gx_zrandom"
  external nrandom1_in :               tref ->       int->                unit   = "gx_nrandom1"
  external zrandom1_in :               tref ->       int->                unit   = "gx_zrandom1"
  external xrandom_init:                             int->                unit   = "gx_random_init"
  external nbits       :                               t ->               int    = "gx_nbits"
  external lowbits     :                               t ->               int    = "gx_lowbits"
  external highbits    :                               t ->               int    = "gx_highbits"
  external nth_word    :                               t -> int ->        int    = "gx_nth_word"
  external nth_bit     :                               t -> int ->        bool   = "gx_nth_bit"
  external shl         :                               t -> int ->        t      = "gx_f_shl"
  external shr         :                               t -> int ->        t      = "gx_f_shr"
  external split       :                               t -> int ->        t*t    = "gx_f_split"
  external join        :                               t -> t   -> int -> t      = "gx_f_join"
  external shl_in      :               tref ->         t -> int ->        unit   = "gx_shl"
  external shr_in      :               tref ->         t -> int ->        unit   = "gx_shr"
  external split_in    :               tref -> tref -> t -> int ->        unit   = "gx_split"
  external join_in     :               tref ->         t -> t   -> int -> unit   = "gx_join"
  external dump        :                               t ->               unit   = "gx_dump"

  let zero   = of_int 0
  let one    = of_int 1

  let random_init x =
    xrandom_init x;
    if x = 0 then Random.self_init() else Random.init(x)

  (* affichage tronqu  1000 caractres *)
  let toplevel_print(a) =
    let s = string_of a      in
    let l = String.length(s) in
    if l < 1000 then Format.print_string s
    else begin
      Format.print_string (String.sub s 0 100);
      Format.print_string " ... (";
      Format.print_int    (l-200);
      Format.print_string " digits) ... ";
      Format.print_string (String.sub s (l-100) 100)
    end

  let toplevel_print_tref(a) =
    Format.print_string "tref(";
    toplevel_print(look a);
    Format.print_char ')'

  exception Error of string
  let _ = Callback.register_exception "gx_error" (Error "")

  external register : unit -> unit = "gx_register"
  let _ = register()

  external isprime     :                               t ->           tristate   = "gx_isprime"
  external isprime_1   :                             int ->           tristate   = "gx_isprime_1"
end

                     (* +-----------------------------+
                        |  Comparaison entre modules  |
                        +-----------------------------+ *)

module Cmp(A:Int_type)(B:Int_type) = struct
  open Printf

  type t    = A.t    * B.t
  type tref = A.tref * B.tref

  let aname = A.name()
  and bname = B.name()
  let name() = sprintf "Cmp(%s,%s)" aname bname

  let zero = (A.zero,B.zero)
  let one  = (A.one, B.one)

  (* conversion entre lments de type A et de type B *)
  let a_of_b(x) = A.of_string(B.hstring_of x)
  let b_of_a(x) = B.of_string(A.hstring_of x)

  (* compare une valeur de A et une valeur de B *)
  let egal a b = (A.hstring_of a) = (B.hstring_of b)

  (* copie d'une valeur suceptible d'tre crase *)
  let acopy a = let ra = A.make_ref a in A.look ra
  let bcopy b = let rb = B.make_ref b in B.look rb

  (* noms des modes d'arrondi *)
  let mstring_of = function
    | Floor        -> "Floor"
    | Ceil         -> "Ceil"
    | Nearest_up   -> "Nearest_up"
    | Nearest_down -> "Nearest_down"

  (* noms des rsultats tristate *)
  let cstring_of = function
    | False   -> "False"
    | Unknown -> "Unknown"
    | True    -> "True"

  (*
    comparaison et agrgation de rsultats calculs dans A et B :
    on compare smantiquement les rsultats et en cas d''accord on
    retourne le couple de ces rsultats (ou un seul si le rsultat
    est un int,string,bool). S'il n'y a pas galit on dclenche
    l''exception Differ.

    abrviations pour les types des arguments/rsultats
    tref       = r
    t          = t
    int        = i
    unit       = u
    round_mode = m
    string     = s
    bool       = b
    tristate   = c
  *)

  exception Error of string

  (* affiche le nom de la fonction et les paramtres/rsultats *)
  let dif_m_t_i_i fn m a1 ia2 ia3 b1 ib2 ib3 =
    sprintf "%s.%s = (%s,%s,%d,%d) %s.%s = (%s,%s,%d,%d)"
      aname fn (mstring_of m) (A.string_of a1) ia2 ia3
      bname fn (mstring_of m) (B.string_of b1) ib2 ib3

  let dif_m_t_i_t fn m a1 ia2 a3 b1 ib2 b3 =
    sprintf "%s.%s = (%s,%s,%d,%s) %s.%s = (%s,%s,%d,%s)"
      aname fn (mstring_of m) (A.string_of a1) ia2 (A.string_of a3)
      bname fn (mstring_of m) (B.string_of b1) ib2 (B.string_of b3)

  let dif_m_t_i_t_i fn m a1 ia2 a3 ia4 b1 ib2 b3 ib4 =
    sprintf "%s.%s = (%s,%s,%d,%s,%d) %s.%s = (%s,%s,%d,%s,%d)"
      aname fn (mstring_of m) (A.string_of a1) ia2 (A.string_of a3) ia4
      bname fn (mstring_of m) (B.string_of b1) ib2 (B.string_of b3) ib4

  let dif_m_t_t fn m a1 a2 b1 b2 =
    sprintf "%s.%s = (%s,%s,%s) %s.%s = (%s,%s,%s)"
      aname fn (mstring_of m) (A.string_of a1) (A.string_of a2)
      bname fn (mstring_of m) (B.string_of b1) (B.string_of b2)

  let dif_m_t_t_t fn m a1 a2 a3 b1 b2 b3 =
    sprintf "%s.%s = (%s,%s,%s,%s) %s.%s = (%s,%s,%s,%s)"
      aname fn (mstring_of m) (A.string_of a1) (A.string_of a2) (A.string_of a3)
      bname fn (mstring_of m) (B.string_of b1) (B.string_of b2) (B.string_of b3)

  let dif_m_t_t_t_t fn m a1 a2 a3 a4 b1 b2 b3 b4 =
    sprintf "%s.%s = (%s,%s,%s,%s,%s) %s.%s = (%s,%s,%s,%s,%s)"
      aname fn (mstring_of m) (A.string_of a1) (A.string_of a2) (A.string_of a3) (A.string_of a4)
      bname fn (mstring_of m) (B.string_of b1) (B.string_of b2) (B.string_of b3) (B.string_of b4)

  let dif_i_c fn ia1 ca2 ib1 cb2 =
    sprintf "%s.%s = (%d,%s) %s.%s = (%d,%s)"
      aname fn ia1 (cstring_of ca2)
      bname fn ib1 (cstring_of cb2)

  let dif_i_i_t fn ia1 ia2 a3 ib1 ib2 b3 =
    sprintf "%s.%s = (%d,%d,%s) %s.%s = (%d,%d,%s)"
      aname fn ia1 ia2 (A.string_of a3)
      bname fn ib1 ib2 (B.string_of b3)

  let dif_i_t fn ia1 a2 ib1 b2 =
    sprintf "%s.%s = (%d,%s) %s.%s = (%d,%s)"
      aname fn ia1 (A.string_of a2)
      bname fn ib1 (B.string_of b2)

  let dif_s fn sa1 sb1 =
    sprintf "%s.%s = (%s) %s.%s = (%s)"
      aname fn sa1
      bname fn sb1

  let dif_s_t fn sa1 a2 sb1 b2 =
    sprintf "%s.%s = (%s,%s) %s.%s = (%s,%s)"
      aname fn sa1 (A.string_of a2)
      bname fn sb1 (B.string_of b2)

  let dif_t_c fn a1 ca2 b1 cb2 =
    sprintf "%s.%s = (%s,%s) %s.%s = (%s,%s)"
      aname fn (A.string_of a1) (cstring_of ca2)
      bname fn (B.string_of b1) (cstring_of cb2)

  let dif_t_i fn a1 ia2 b1 ib2 =
    sprintf "%s.%s = (%s,%d) %s.%s = (%s,%d)"
      aname fn (A.string_of a1) ia2
      bname fn (B.string_of b1) ib2

  let dif_t_i_b fn a1 ia2 ba3 b1 ib2 bb3 =
    sprintf "%s.%s = (%s,%d,%b) %s.%s = (%s,%d,%b)"
      aname fn (A.string_of a1) ia2 ba3
      bname fn (B.string_of b1) ib2 bb3

  let dif_t_i_i fn a1 ia2 ia3 b1 ib2 ib3 =
    sprintf "%s.%s = (%s,%d,%d) %s.%s = (%s,%d,%d)"
      aname fn (A.string_of a1) ia2 ia3
      bname fn (B.string_of b1) ib2 ib3

  let dif_t_i_t fn a1 ia2 a3 b1 ib2 b3 =
    sprintf "%s.%s = (%s,%d,%s) %s.%s = (%s,%d,%s)"
      aname fn (A.string_of a1) ia2 (A.string_of a3)
      bname fn (B.string_of b1) ib2 (B.string_of b3)

  let dif_t_i_t_i fn a1 ia2 a3 ia4 b1 ib2 b3 ib4 =
    sprintf "%s.%s = (%s,%d,%s,%d) %s.%s = (%s,%d,%s,%d)"
      aname fn (A.string_of a1) ia2 (A.string_of a3) ia4
      bname fn (B.string_of b1) ib2 (B.string_of b3) ib4

  let dif_t_i_t_t fn a1 ia2 a3 a4 b1 ib2 b3 b4 =
    sprintf "%s.%s = (%s,%d,%s,%s) %s.%s = (%s,%d,%s,%s)"
      aname fn (A.string_of a1) ia2 (A.string_of a3) (A.string_of a4)
      bname fn (B.string_of b1) ib2 (B.string_of b3) (B.string_of b4)

  let dif_t_s fn a1 sa2 b1 sb2 =
    sprintf "%s.%s = (%s,%s) %s.%s = (%s,%s)"
      aname fn (A.string_of a1) sa2
      bname fn (B.string_of b1) sb2

  let dif_t_t fn a1 a2 b1 b2 =
    sprintf "%s.%s = (%s,%s) %s.%s = (%s,%s)"
      aname fn (A.string_of a1) (A.string_of a2)
      bname fn (B.string_of b1) (B.string_of b2)

  let dif_t_t_b fn a1 a2 ba3 b1 b2 bb3 =
    sprintf "%s.%s = (%s,%s,%b) %s.%s = (%s,%s,%b)"
      aname fn (A.string_of a1) (A.string_of a2) ba3
      bname fn (B.string_of b1) (B.string_of b2) bb3

  let dif_t_t_i fn a1 a2 ia3 b1 b2 ib3 =
    sprintf "%s.%s = (%s,%s,%d) %s.%s = (%s,%s,%d)"
      aname fn (A.string_of a1) (A.string_of a2) ia3
      bname fn (B.string_of b1) (B.string_of b2) ib3

  let dif_t_t_i_t fn a1 a2 ia3 a4 b1 b2 ib3 b4 =
    sprintf "%s.%s = (%s,%s,%d,%s) %s.%s = (%s,%s,%d,%s)"
      aname fn (A.string_of a1) (A.string_of a2) ia3 (A.string_of a4)
      bname fn (B.string_of b1) (B.string_of b2) ib3 (B.string_of b4)

  let dif_t_t_t fn a1 a2 a3 b1 b2 b3 =
    sprintf "%s.%s = (%s,%s,%s) %s.%s = (%s,%s,%s)"
      aname fn (A.string_of a1) (A.string_of a2) (A.string_of a3)
      bname fn (B.string_of b1) (B.string_of b2) (B.string_of b3)

  let dif_t_t_t_t fn a1 a2 a3 a4 b1 b2 b3 b4 =
    sprintf "%s.%s = (%s,%s,%s,%s) %s.%s = (%s,%s,%s,%s)"
      aname fn (A.string_of a1) (A.string_of a2) (A.string_of a3) (A.string_of a4)
      bname fn (B.string_of b1) (B.string_of b2) (B.string_of b3) (B.string_of b4)

  let dif_t_t_t_t_t fn a1 a2 a3 a4 a5 b1 b2 b3 b4 b5 =
    sprintf "%s.%s = (%s,%s,%s,%s,%s) %s.%s = (%s,%s,%s,%s,%s)"
      aname fn (A.string_of a1) (A.string_of a2) (A.string_of a3) (A.string_of a4) (A.string_of a5)
      bname fn (B.string_of b1) (B.string_of b2) (B.string_of b3) (B.string_of b4) (B.string_of b5)

  let dif_t_t_t_t_t_t_t fn a1 a2 a3 a4 a5 a6 a7 b1 b2 b3 b4 b5 b6 b7 =
    sprintf "%s.%s = (%s,%s,%s,%s,%s,%s,%s) %s.%s = (%s,%s,%s,%s,%s,%s,%s)"
      aname fn (A.string_of a1) (A.string_of a2) (A.string_of a3) (A.string_of a4) (A.string_of a5) (A.string_of a6) (A.string_of a7)
      bname fn (B.string_of b1) (B.string_of b2) (B.string_of b3) (B.string_of b4) (B.string_of b5) (B.string_of b6) (B.string_of b7)

  (*
    construction des fonctions de comparaison :
     x_y_z fa fb fn prend les fonctions fa (module A), fb (module B),
     de signatures x -> y -> z, et retourne une fonction de mme signature
     qui effectue les calculs en double puis compare les rsultats.

     Les contenus des rfrences sont compars en sortie seulement
     Les paramtres qui risquent d'tre crass en cas d'opration sur place
     sont sauvegards avant calcul
  *)
  let i_i_t fa fb fn = fun i1 i2 ->
    let a3 = fa i1 i2
    and b3 = fb i1 i2 in
    if egal a3 b3 then (a3,b3)
    else raise (Error (dif_i_i_t fn i1 i2 a3 i1 i2 b3))

  let i_t fa fb fn = fun i1 ->
    let a2 = fa i1
    and b2 = fb i1 in
    if egal a2 b2 then (a2,b2)
    else raise (Error (dif_i_t fn i1 a2 i1 b2))

  let i_u fa fb fn = fun i1 ->
    fa i1;
    fb i1

  let m_r_r_t_t_u fa fb fn = fun m (ra1,rb1) (ra2,rb2) (a3,b3) (a4,b4) ->
    let aa3 = acopy a3 and aa4 = acopy a4
    and bb3 = bcopy b3 and bb4 = bcopy b4 in
    fa m ra1 ra2 a3 a4;
    fb m rb1 rb2 b3 b4;
    let a1 = A.look ra1 and a2 = A.look ra2
    and b1 = B.look rb1 and b2 = B.look rb2 in
    if (egal a1 b1) & (egal a2 b2) then ()
    else raise (Error(dif_m_t_t_t_t fn m aa3 aa4 a1 a2 bb3 bb4 b1 b2))

  let m_r_t_i_i fa fb fn = fun m (ra1,rb1) (a2,b2) i3 ->
    let aa2 = acopy a2
    and bb2 = bcopy b2 in
    let ia4 = fa m ra1 a2 i3 and a1 = A.look ra1
    and ib4 = fb m rb1 b2 i3 and b1 = B.look rb1 in
    if (egal a1 b1) & (ia4 = ib4) then ia4
    else raise (Error (dif_m_t_i_t_i fn m aa2 i3 a1 ia4 bb2 i3 b1 ib4))

  let m_r_t_i_u fa fb fn = fun m (ra1,rb1) (a2,b2) i3 ->
    let aa2 = acopy a2
    and bb2 = bcopy b2 in
    fa m ra1 a2 i3; let a1 = A.look ra1 in
    fb m rb1 b2 i3; let b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_m_t_i_t fn m aa2 i3 a1 bb2 i3 b1))

  let m_r_t_t_t_u fa fb fn = fun m (ra1,rb1) (a2,b2) (a3,b3) (a4,b4) ->
    let aa2 = acopy a2 and aa3 = acopy a3 and aa4 = acopy a4
    and bb2 = bcopy b2 and bb3 = bcopy b3 and bb4 = bcopy b4 in
    fa m ra1 a2 a3 a4;
    fb m rb1 b2 b3 b4;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_m_t_t_t_t fn m aa2 aa3 aa4 a1 bb2 bb3 bb4 b1))

  let m_r_t_t_u fa fb fn = fun m (ra1,rb1) (a2,b2) (a3,b3) ->
    let aa2 = acopy a2 and aa3 = acopy a3 
    and bb2 = bcopy b2 and bb3 = bcopy b3 in
    fa m ra1 a2 a3;
    fb m rb1 b2 b3;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_m_t_t_t fn m aa2 aa3 a1 bb2 bb3 b1))

  let m_r_t_u fa fb fn = fun m (ra1,rb1) (a2,b2) ->
    let aa2 = acopy a2 
    and bb2 = bcopy b2 in
    fa m ra1 a2;
    fb m rb1 b2;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_m_t_t fn m aa2 a1 bb2 b1))

  let m_t_i_i fa fb fn = fun m (a1,b1) i2 ->
    let ia3 = fa m a1 i2
    and ib3 = fb m b1 i2 in
    if ia3 = ib3 then ia3
    else raise (Error (dif_m_t_i_i fn m a1 i2 ia3 b1 i2 ib3))

  let m_t_i_t fa fb fn = fun m (a1,b1) i2 ->
    let a3 = fa m a1 i2
    and b3 = fb m b1 i2 in
    if egal a3 b3 then (a3,b3)
    else raise (Error (dif_m_t_i_t fn m a1 i2 a3 b1 i2 b3))

  let m_t_i_ti fa fb fn = fun m (a1,b1) i2 ->
    let a3,ia4 = fa m a1 i2
    and b3,ib4 = fb m b1 i2 in
    if (egal a3 b3) & (ia4 = ib4)  then (a3,b3),ia4
    else raise (Error (dif_m_t_i_t_i fn m a1 i2 a3 ia4 b1 i2 b3 ib4))

  let m_t_t fa fb fn = fun m (a1,b1) ->
    let a2 = fa m a1
    and b2 = fb m b1 in
    if egal a2  b2 then (a2,b2)
    else raise (Error (dif_m_t_t fn m a1 a2 b1 b2))

  let m_t_t_t fa fb fn = fun m (a1,b1) (a2,b2) ->
    let a3 = fa m a1 a2
    and b3 = fb m b1 b2 in
    if egal a3  b3 then (a3,b3)
    else raise (Error (dif_m_t_t_t fn m a1 a2 a3 b1 b2 b3))

  let m_t_t_t_t fa fb fn = fun m (a1,b1) (a2,b2) (a3,b3) ->
    let a4 = fa m a1 a2 a3
    and b4 = fb m b1 b2 b3 in
    if egal a4  b4 then (a4,b4)
    else raise (Error (dif_m_t_t_t_t fn m a1 a2 a3 a4 b1 b2 b3 b4))

  let m_t_t_tt fa fb fn = fun m (a1,b1) (a2,b2) ->
    let a3,a4 = fa m a1 a2
    and b3,b4 = fb m b1 b2 in
    if (egal a3  b3) & (egal a4 b4) then (a3,b3),(a4,b4)
    else raise (Error (dif_m_t_t_t_t fn m a1 a2 a3 a4 b1 b2 b3 b4))

  let r_i_i_u fa fb fn = fun (ra1,rb1) i2 i3 ->
    fa ra1 i2 i3;
    fb rb1 i2 i3;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_i_i_t fn i2 i3 a1 i2 i3 b1))

  let r_i_u fa fb fn = fun (ra1,rb1) i2 ->
    fa ra1 i2;
    fb rb1 i2;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_i_t fn i2 a1 i2 b1))

  let r_r_r_r_r_t_t_u fa fb fn = fun (ra1,rb1) (ra2,rb2) (ra3,rb3) (ra4,rb4) (ra5,rb5) (a6,b6) (a7,b7) ->
    let aa6 = acopy a6 and aa7 = acopy a7
    and bb6 = bcopy b6 and bb7 = bcopy b7 in
    fa ra1 ra2 ra3 ra4 ra5 a6 a7;
    fb rb1 rb2 rb3 rb4 rb5 b6 b7;
    let a1 = A.look ra1 and a2 = A.look ra2 and a3 = A.look ra3 and a4 = A.look ra4 and a5 = A.look ra5
    and b1 = B.look rb1 and b2 = B.look rb2 and b3 = B.look rb3 and b4 = B.look rb4 and b5 = B.look rb5 in
    if (egal a1 b1) & (egal a2 b2) & (egal a3 b3) & (egal a4 b4) & (egal a5 b5) then ()
    else raise (Error (dif_t_t_t_t_t_t_t fn aa6 aa7 a1 a2 a3 a4 a5 bb6 bb7 b1 b2 b3 b4 b5))

  let r_r_r_t_t_u fa fb fn = fun (ra1,rb1) (ra2,rb2) (ra3,rb3) (a4,b4) (a5,b5) ->
    let aa4 = acopy a4 and aa5 = acopy a5
    and bb4 = bcopy b4 and bb5 = bcopy b5 in
    fa ra1 ra2 ra3 a4 a5;
    fb rb1 rb2 rb3 b4 b5;
    let a1 = A.look ra1 and a2 = A.look ra2 and a3 = A.look ra3
    and b1 = B.look rb1 and b2 = B.look rb2 and b3 = B.look rb3 in
    if (egal a1 b1) & (egal a2 b2) & (egal a3 b3) then ()
    else raise (Error (dif_t_t_t_t_t fn aa4 aa5 a1 a2 a3 bb4 bb5 b1 b2 b3))

  let r_r_t_i_u fa fb fn = fun (ra1,rb1) (ra2,rb2) (a3,b3) i4 ->
    let aa3 = acopy a3
    and bb3 = bcopy b3 in
    fa ra1 ra2 a3 i4;
    fb rb1 rb2 b3 i4;
    let a1 = A.look ra1 and a2 = A.look ra2
    and b1 = B.look rb1 and b2 = B.look rb2 in
    if (egal a1 b1) & (egal a2 b2) then ()
    else raise (Error (dif_t_i_t_t fn aa3 i4 a1 a2 bb3 i4 b1 b2))

  let r_r_t_t_u fa fb fn = fun (ra1,rb1) (ra2,rb2) (a3,b3) (a4,b4) ->
    let aa3 = acopy a3 and aa4 = acopy a4
    and bb3 = bcopy b3 and bb4 = bcopy b4 in
    fa ra1 ra2 a3 a4;
    fb rb1 rb2 b3 b4;
    let a1 = A.look ra1 and a2 = A.look ra2
    and b1 = B.look rb1 and b2 = B.look rb2 in
    if (egal a1 b1) & (egal a2 b2) then ()
    else raise (Error (dif_t_t_t_t fn aa3 aa4 a1 a2 bb3 bb4 b1 b2))

  let r_s_u fa fb fn = fun (ra1,rb1) s2 ->
    fa ra1 s2;
    fb rb1 s2;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_s_t fn s2 a1 s2 b1))

  let r_t fa fb fn = fun (ra1,rb1) ->
    let a2 = A.look ra1
    and b2 = B.look rb1 in (a2,b2)

  let r_t_i_i fa fb fn = fun (ra1,rb1) (a2,b2) i3 ->
    let aa2 = acopy a2
    and bb2 = bcopy b2 in
    let ia4 = fa ra1 a2 i3
    and ib4 = fb rb1 b2 i3 in
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if (egal a1 b1) & (ia4 = ib4) then ia4
    else raise (Error (dif_t_i_t_i fn aa2 i3 a1 ia4 bb2 i3 b1 ib4))

  let r_t_i_u fa fb fn = fun (ra1,rb1) (a2,b2) i3 ->
    let aa2 = acopy a2
    and bb2 = bcopy b2 in
    fa ra1 a2 i3;
    fb rb1 b2 i3;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_t_i_t fn aa2 i3 a1 bb2 i3 b1))

  let r_t_t_i_u fa fb fn = fun (ra1,rb1) (a2,b2) (a3,b3) i4 ->
    let aa2 = acopy a2 and aa3 = acopy a3
    and bb2 = bcopy b2 and bb3 = bcopy b3 in
    fa ra1 a2 a3 i4;
    fb rb1 b2 b3 i4;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_t_t_i_t fn aa2 aa3 i4 a1 bb2 bb3 i4 b1))

  let r_t_t_t_u fa fb fn = fun (ra1,rb1) (a2,b2) (a3,b3) (a4,b4) ->
    let aa2 = acopy a2 and aa3 = acopy a3 and aa4 = acopy a4
    and bb2 = bcopy b2 and bb3 = bcopy b3 and bb4 = bcopy b4 in
    fa ra1 a2 a3 a4;
    fb rb1 b2 b3 b4;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_t_t_t_t fn aa2 aa3 aa4 a1 bb2 bb3 bb4 b1))

  let r_t_t_u fa fb fn = fun (ra1,rb1) (a2,b2) (a3,b3) ->
    let aa2 = acopy a2 and aa3 = acopy a3
    and bb2 = bcopy b2 and bb3 = bcopy b3 in
    fa ra1 a2 a3;
    fb rb1 b2 b3;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_t_t_t fn aa2 aa3 a1 bb2 bb3 b1))

  let r_t_u fa fb fn = fun (ra1,rb1) (a2,b2) ->
    let aa2 = acopy a2
    and bb2 = bcopy b2 in
    fa ra1 a2;
    fb rb1 b2;
    let a1 = A.look ra1
    and b1 = B.look rb1 in
    if egal a1 b1 then ()
    else raise (Error (dif_t_t fn aa2 a1 bb2 b1))

  let s_t fa fb fn = fun s1 ->
    let a2 = fa s1
    and b2 = fb s1 in
    if egal a2 b2 then (a2,b2)
    else raise (Error (dif_s_t fn s1 a2 s1 b2))

  let t_i fa fb fn = fun (a1,b1) ->
    let ia2 = fa a1
    and ib2 = fb b1 in
    if ia2 = ib2 then ia2
    else raise (Error (dif_t_i fn a1 ia2 b1 ib2))

  let t_i_b fa fb fn = fun (a1,b1) i2 ->
    let ba3 = fa a1 i2
    and bb3 = fb b1 i2 in
    if ba3 = bb3 then ba3
    else raise (Error (dif_t_i_b fn a1 i2 ba3 b1 i2 bb3))

  let t_i_i fa fb fn = fun (a1,b1) i2 ->
    let ia3 = fa a1 i2
    and ib3 = fb b1 i2 in
    if ia3 = ib3 then ia3
    else raise (Error (dif_t_i_i fn a1 i2 ia3 b1 i2 ib3))

  let t_i_t fa fb fn = fun (a1,b1) i2 ->
    let a3 = fa a1 i2
    and b3 = fb b1 i2 in
    if egal a3  b3 then (a3,b3)
    else raise (Error (dif_t_i_t fn a1 i2 a3 b1 i2 b3))

  let t_i_ti fa fb fn = fun (a1,b1) i2 ->
    let a3,ia4 = fa a1 i2
    and b3,ib4 = fb b1 i2 in
    if (egal a3 b3) & (ia4 = ib4) then (a3,b3),ia4
    else raise (Error (dif_t_i_t_i fn a1 i2 a3 ia4 b1 i2 b3 ib4))

  let t_i_tt fa fb fn = fun (a1,b1) i2 ->
    let a3,a4 = fa a1 i2
    and b3,b4 = fb b1 i2 in
    if (egal a3 b3) & (egal a4 b4) then (a3,b3),(a4,b4)
    else raise (Error (dif_t_i_t_t fn a1 i2 a3 a4 b1 i2 b3 b4))

  let t_r fa fb fn = fun (a1,b1) ->
    let ra2 = fa a1
    and rb2 = fb b1 in (ra2,rb2)

  let t_s fa fb fn = fun (a1,b1) ->
    let sa2 = fa a1
    and sb2 = fb b1 in
    if sa2 = sb2 then sa2
    else raise (Error (dif_t_s fn a1 sa2 b1 sb2))

  let t_t fa fb fn = fun (a1,b1) ->
    let a2 = fa a1
    and b2 = fb b1 in
    if egal a2 b2 then (a2,b2)
    else raise (Error (dif_t_t fn a1 a2 b1 b2))

  let t_t_b fa fb fn = fun (a1,b1) (a2,b2) ->
    let ba3 = fa a1 a2
    and bb3 = fb b1 b2 in
    if ba3 = bb3 then ba3
    else raise (Error (dif_t_t_b fn a1 a2 ba3 b1 b2 bb3))

  let t_t_i fa fb fn = fun (a1,b1) (a2,b2) ->
    let ia3 = fa a1 a2
    and ib3 = fb b1 b2 in
    if ia3 = ib3 then ia3
    else raise (Error (dif_t_t_i fn a1 a2 ia3 b1 b2 ib3))

  let t_t_i_t fa fb fn = fun (a1,b1) (a2,b2) i3 ->
    let a4 = fa a1 a2 i3
    and b4 = fb b1 b2 i3 in
    if egal a4 b4 then (a4,b4)
    else raise (Error (dif_t_t_i_t fn a1 a2 i3 a4 b1 b2 i3 b4))

  let t_t_t fa fb fn = fun (a1,b1) (a2,b2) ->
    let a3 = fa a1 a2
    and b3 = fb b1 b2 in
    if egal a3 b3 then (a3,b3)
    else raise (Error (dif_t_t_t fn a1 a2 a3 b1 b2 b3))

  let t_t_t_t fa fb fn = fun (a1,b1) (a2,b2) (a3,b3) ->
    let a4 = fa a1 a2 a3
    and b4 = fb b1 b2 b3 in
    if egal a4 b4 then (a4,b4)
    else raise (Error (dif_t_t_t_t fn a1 a2 a3 a4 b1 b2 b3 b4))

  let t_t_tt fa fb fn = fun (a1,b1) (a2,b2) ->
    let a3,a4 = fa a1 a2
    and b3,b4 = fb b1 b2 in
    if (egal a3 b3) & (egal a4 b4) then (a3,b3),(a4,b4)
    else raise (Error (dif_t_t_t_t fn a1 a2 a3 a4 b1 b2 b3 b4))

  let t_t_ttt fa fb fn = fun (a1,b1) (a2,b2) ->
    let a3,a4,a5 = fa a1 a2
    and b3,b4,b5 = fb b1 b2 in
    if (egal a3 b3) & (egal a4 b4) & (egal a5 b5) then (a3,b3),(a4,b4),(a5,b5)
    else raise (Error (dif_t_t_t_t_t fn a1 a2 a3 a4 a5 b1 b2 b3 b4 b5))

  let t_t_ttttt fa fb fn = fun (a1,b1) (a2,b2) ->
    let a3,a4,a5,a6,a7 = fa a1 a2
    and b3,b4,b5,b6,b7 = fb b1 b2 in
    if (egal a3 b3) & (egal a4 b4) & (egal a5 b5) & (egal a6 b6) & (egal a7 b7) then (a3,b3),(a4,b4),(a5,b5),(a6,b6),(a7,b7)
    else raise (Error (dif_t_t_t_t_t_t_t fn a1 a2 a3 a4 a5 a6 a7 b1 b2 b3 b4 b5 b6 b7))


  (* oprations  implmenter *)
  let make_ref    = t_r             A.make_ref     B.make_ref     "make_ref"
  let copy_in     = r_t_u           A.copy_in      B.copy_in      "copy_in"
  let copy_out    = r_t             A.copy_out     B.copy_out     "copy_out"
  let look        = r_t             A.look         B.look         "look"
  let add         = t_t_t           A.add          B.add          "add"
  let add_1       = t_i_t           A.add_1        B.add_1        "add_1"
  let add_in      = r_t_t_u         A.add_in       B.add_in       "add_in"
  let add_1_in    = r_t_i_u         A.add_1_in     B.add_1_in     "add_1_in"
  let sub         = t_t_t           A.sub          B.sub          "sub"
  let sub_1       = t_i_t           A.sub_1        B.sub_1        "sub_1"
  let sub_in      = r_t_t_u         A.sub_in       B.sub_in       "sub_in"
  let sub_1_in    = r_t_i_u         A.sub_1_in     B.sub_1_in     "sub_1_in"
  let mul         = t_t_t           A.mul          B.mul          "mul"
  let mul_1       = t_i_t           A.mul_1        B.mul_1        "mul_1"
  let mul_in      = r_t_t_u         A.mul_in       B.mul_in       "mul_in"
  let mul_1_in    = r_t_i_u         A.mul_1_in     B.mul_1_in     "mul_1_in"
  let quomod      = t_t_tt          A.quomod       B.quomod       "quomod"
  let quo         = t_t_t           A.quo          B.quo          "quo"
  let modulo      = t_t_t           A.modulo       B.modulo       "modulo"
  let gquomod     = m_t_t_tt        A.gquomod      B.gquomod      "gquomod"
  let gquo        = m_t_t_t         A.gquo         B.gquo         "gquo"
  let gmod        = m_t_t_t         A.gmod         B.gmod         "gmod"
  let quomod_in   = r_r_t_t_u       A.quomod_in    B.quomod_in    "quomod_in"
  let quo_in      = r_t_t_u         A.quo_in       B.quo_in       "quo_in"
  let mod_in      = r_t_t_u         A.mod_in       B.mod_in       "mod_in"
  let gquomod_in  = m_r_r_t_t_u     A.gquomod_in   B.gquomod_in   "gquomod_in"
  let gquo_in     = m_r_t_t_u       A.gquo_in      B.gquo_in      "gquo_in"
  let gmod_in     = m_r_t_t_u       A.gmod_in      B.gmod_in      "gmod_in"
  let quomod_1    = t_i_ti          A.quomod_1     B.quomod_1     "quomod_1"
  let quo_1       = t_i_t           A.quo_1        B.quo_1        "quo_1"
  let mod_1       = t_i_i           A.mod_1        B.mod_1        "mod_1"
  let gquomod_1   = m_t_i_ti        A.gquomod_1    B.gquomod_1    "gquomod_1"
  let gquo_1      = m_t_i_t         A.gquo_1       B.gquo_1       "gquo_1"
  let gmod_1      = m_t_i_i         A.gmod_1       B.gmod_1       "gmod_1"
  let quomod_1_in = r_t_i_i         A.quomod_1_in  B.quomod_1_in  "quomod_1_in"
  let quo_1_in    = r_t_i_u         A.quo_1_in     B.quo_1_in     "quo_1_in"
  let gquomod_1_in= m_r_t_i_i       A.gquomod_1_in B.gquomod_1_in "gquomod_1_in"
  let gquo_1_in   = m_r_t_i_u       A.gquo_1_in    B.gquo_1_in    "gquo_1_in"
  let abs         = t_t             A.abs          B.abs          "abs"
  let abs_in      = r_t_u           A.abs_in       B.abs_in       "abs_in"
  let neg         = t_t             A.neg          B.neg          "neg"
  let neg_in      = r_t_u           A.neg_in       B.neg_in       "neg_in"
  let sqr         = t_t             A.sqr          B.sqr          "sqr"
  let pow         = t_i_t           A.pow          B.pow          "pow"
  let pow_1       = i_i_t           A.pow_1        B.pow_1        "pow_1"
  let powmod      = t_t_t_t         A.powmod       B.powmod       "powmod"
  let gpowmod     = m_t_t_t_t       A.gpowmod      B.gpowmod      "gpowmod"
  let sqr_in      = r_t_u           A.sqr_in       B.sqr_in       "sqr_in"
  let pow_in      = r_t_i_u         A.pow_in       B.pow_in       "pow_in"
  let pow_1_in    = r_i_i_u         A.pow_1_in     B.pow_1_in     "pow_1_in"
  let powmod_in   = r_t_t_t_u       A.powmod_in    B.powmod_in    "powmod_in"
  let gpowmod_in  = m_r_t_t_t_u     A.gpowmod_in   B.gpowmod_in   "gpowmod_in"
  let sqrt        = t_t             A.sqrt         B.sqrt         "sqrt"
  let root        = t_i_t           A.root         B.root         "root"
  let gsqrt       = m_t_t           A.gsqrt        B.gsqrt        "gsqrt"
  let groot       = m_t_i_t         A.groot        B.groot        "groot"
  let sqrt_in     = r_t_u           A.sqrt_in      B.sqrt_in      "sqrt_in"
  let root_in     = r_t_i_u         A.root_in      B.root_in      "root_in"
  let gsqrt_in    = m_r_t_u         A.gsqrt_in     B.gsqrt_in     "gsqrt_in"
  let groot_in    = m_r_t_i_u       A.groot_in     B.groot_in     "groot_in"
  let fact        = i_t             A.fact         B.fact         "fact"
  let fact_in     = r_i_u           A.fact_in      B.fact_in      "fact_in"
  let gcd         = t_t_t           A.gcd          B.gcd          "gcd"
(*let gcd_ex      = t_t_ttt         A.gcd_ex       B.gcd_ex       "gcd_ex"   *)
(*let cfrac       = t_t_ttttt       A.cfrac        B.cfrac        "cfrac"    *)
  let gcd_in      = r_t_t_u         A.gcd_in       B.gcd_in       "gcd_in"
(*let gcd_ex_in   = r_r_r_t_t_u     A.gcd_ex_in    B.gcd_ex_in    "gcd_ex_in"*)
(*let cfrac_in    = r_r_r_r_r_t_t_u A.cfrac_in     B.cfrac_in     "cfrac_in" *)
  let sgn         = t_i             A.sgn          B.sgn          "sgn"
  let cmp         = t_t_i           A.cmp          B.cmp          "cmp"
  let cmp_1       = t_i_i           A.cmp_1        B.cmp_1        "cmp_1"
  let eq          = t_t_b           A.eq           B.eq           "eq"
  let eq_1        = t_i_b           A.eq_1         B.eq_1         "eq_1"
  let neq         = t_t_b           A.neq          B.neq          "neq"
  let neq_1       = t_i_b           A.neq_1        B.neq_1        "neq_1"
  let inf         = t_t_b           A.inf          B.inf          "inf"
  let inf_1       = t_i_b           A.inf_1        B.inf_1        "inf_1"
  let infeq       = t_t_b           A.infeq        B.infeq        "infeq"
  let infeq_1     = t_i_b           A.infeq_1      B.infeq_1      "infeq_1"
  let sup         = t_t_b           A.sup          B.sup          "sup"
  let sup_1       = t_i_b           A.sup_1        B.sup_1        "sup_1"
  let supeq       = t_t_b           A.supeq        B.supeq        "supeq"
  let supeq_1     = t_i_b           A.supeq_1      B.supeq_1      "supeq_1"
  let of_int      = i_t             A.of_int       B.of_int       "of_int"
  let of_string   = s_t             A.of_string    B.of_string    "of_string"
  let of_int_in   = r_i_u           A.of_int_in    B.of_int_in    "of_int_in"
  let of_string_in= r_s_u           A.of_string_in B.of_string_in "of_string_in"
  let int_of      = t_i             A.int_of       B.int_of       "int_of"
  let string_of   = t_s             A.string_of    B.string_of    "string_of"
  let bstring_of  = t_s             A.bstring_of   B.bstring_of   "bstring_of"
  let hstring_of  = t_s             A.hstring_of   B.hstring_of   "hstring_of"
  let ostring_of  = t_s             A.ostring_of   B.ostring_of   "ostring_of"
(*let nrandom     = i_t             A.nrandom      B.nrandom      "nrandom"    *)
(*let zrandom     = i_t             A.zrandom      B.zrandom      "zrandom"    *)
(*let nrandom1    = i_t             A.nrandom1     B.nrandom1     "nrandom1"   *)
(*let zrandom1    = i_t             A.zrandom1     B.zrandom1     "zrandom1"   *)
(*let nrandom_in  = r_i_u           A.nrandom_in   B.nrandom_in   "nrandom_in" *)
(*let zrandom_in  = r_i_u           A.zrandom_in   B.zrandom_in   "zrandom_in" *)
(*let nrandom1_in = r_i_u           A.nrandom1_in  B.nrandom1_in  "nrandom1_in"*)
(*let zrandom1_in = r_i_u           A.zrandom1_in  B.zrandom1_in  "zrandom1_in"*)
(*let random_init = i_u             A.random_init  B.random_init  "random_init"*)
  let nbits       = t_i             A.nbits        B.nbits        "nbits"
  let lowbits     = t_i             A.lowbits      B.lowbits      "lowbits"
  let highbits    = t_i             A.highbits     B.highbits     "highbits"
  let nth_word    = t_i_i           A.nth_word     B.nth_word     "nth_word"
  let nth_bit     = t_i_b           A.nth_bit      B.nth_bit      "nth_bit"
  let shl         = t_i_t           A.shl          B.shl          "shl"
  let shr         = t_i_t           A.shr          B.shr          "shr"
  let split       = t_i_tt          A.split        B.split        "split"
  let join        = t_t_i_t         A.join         B.join         "join"
  let shl_in      = r_t_i_u         A.shl_in       B.shl_in       "shl_in"
  let shr_in      = r_t_i_u         A.shr_in       B.shr_in       "shr_in"
  let split_in    = r_r_t_i_u       A.split_in     B.split_in     "split_in"
  let join_in     = r_t_t_i_u       A.join_in      B.join_in      "join_in"

  (* Coefficients de Bzout : on ne compare pas u,v mais on vrifie que ua-vb = d *)
  let gcd_ex (a1,b1) (a2,b2) =
    let a3,a4,a5 = A.gcd_ex a1 a2
    and b3,b4,b5 = B.gcd_ex b1 b2 in
    if (egal a3 b3) & (B.eq b3 (B.sub (B.mul b4 b1) (B.mul b5 b2)))
    then (a3,b3),(a4,(b_of_a a4)),(a5,(b_of_a a5))
    else raise (Error (dif_t_t_t_t_t "gcd_ex" a1 a2 a3 a4 a5 b1 b2 b3 b4 b5))

  let cfrac (a1,b1) (a2,b2) =
    let a3,a4,a5,a6,a7 = A.cfrac a1 a2
    and b3,b4,b5,b6,b7 = B.cfrac b1 b2 in
    if (egal a3 b3) & (B.eq b3 (B.sub (B.mul b4 b1) (B.mul b5 b2))) & (egal a6 b6) & (egal a7 b7)
    then (a3,b3),(a4,(b_of_a a4)),(a5,(b_of_a a5)),(a6,b6),(a7,b7)
    else raise (Error (dif_t_t_t_t_t_t_t "cfrac" a1 a2 a3 a4 a5 a6 a7 b1 b2 b3 b4 b5 b6 b7))

  let gcd_ex_in (ra1,rb1) (ra2,rb2) (ra3,rb3) (a4,b4) (a5,b5) =
    let aa4 = acopy a4 and aa5 = acopy a5
    and bb4 = bcopy b4 and bb5 = bcopy b5 in
    A.gcd_ex_in ra1 ra2 ra3 a4 a5;
    B.gcd_ex_in rb1 rb2 rb3 b4 b5;
    let a1 = A.look ra1 and a2 = A.look ra2 and a3 = A.look ra3
    and b1 = B.look rb1 and b2 = B.look rb2 and b3 = B.look rb3 in
    if (egal a1 b1) & (B.eq b1 (B.sub (B.mul b2 bb4) (B.mul b3 bb5)))
    then begin
      B.copy_in rb2 (b_of_a a2);
      B.copy_in rb3 (b_of_a a3)
    end
    else raise (Error (dif_t_t_t_t_t "gcd_ex_in" aa4 aa5 a1 a2 a3 bb4 bb5 b1 b2 b3))

  let cfrac_in (ra1,rb1) (ra2,rb2) (ra3,rb3) (ra4,rb4) (ra5,rb5) (a6,b6) (a7,b7) =
    let aa6 = acopy a6 and aa7 = acopy a7
    and bb6 = bcopy b6 and bb7 = bcopy b7 in
    A.cfrac_in ra1 ra2 ra3 ra4 ra5 a6 a7;
    B.cfrac_in rb1 rb2 rb3 rb4 rb5 b6 b7;
    let a1 = A.look ra1 and a2 = A.look ra2 and a3 = A.look ra3 and a4 = A.look ra4 and a5 = A.look ra5
    and b1 = B.look rb1 and b2 = B.look rb2 and b3 = B.look rb3 and b4 = B.look rb4 and b5 = B.look rb5 in
    if (egal a1 b1) & (B.eq b1 (B.sub (B.mul b2 bb6) (B.mul b3 bb7))) & (egal a4 b4) & (egal a5 b5)
    then begin
      B.copy_in rb2 (b_of_a a2);
      B.copy_in rb3 (b_of_a a3)
    end
    else raise (Error (dif_t_t_t_t_t_t_t "cfrac_in" aa6 aa7 a1 a2 a3 a4 a5 bb6 bb7 b1 b2 b3 b4 b5))

  (* test de primalit : on prend le rsultat le plus prcis *)
  let isprime (a,b) = match (A.isprime a),(B.isprime b) with
  | Unknown,r -> r
  | r,Unknown -> r
  | r,s       -> if r = s then r else raise (Error (dif_t_c "isprime" a r b s))

  let isprime_1 i = match (A.isprime_1 i),(B.isprime_1 i) with
  | Unknown,r -> r
  | r,Unknown -> r
  | r,s       -> if r = s then r else raise (Error (dif_i_c "isprime_1" i r i s))

  (* nombres alatoires : on prend le gnrateur de A *)
  let nrandom  n = let a = A.nrandom  n in a,(b_of_a a)
  let zrandom  n = let a = A.zrandom  n in a,(b_of_a a)
  let nrandom1 n = let a = A.nrandom1 n in a,(b_of_a a)
  let zrandom1 n = let a = A.zrandom1 n in a,(b_of_a a)

  let nrandom_in  (ra,rb) n = A.nrandom_in  ra n; B.copy_in rb (b_of_a (A.look ra))
  let zrandom_in  (ra,rb) n = A.zrandom_in  ra n; B.copy_in rb (b_of_a (A.look ra))
  let nrandom1_in (ra,rb) n = A.nrandom1_in ra n; B.copy_in rb (b_of_a (A.look ra))
  let zrandom1_in (ra,rb) n = A.zrandom1_in ra n; B.copy_in rb (b_of_a (A.look ra))

  let random_init n = A.random_init n

  (* affichage *)
  let toplevel_print(a,b)      = A.toplevel_print(a)
  let toplevel_print_tref(a,b) = A.toplevel_print_tref(a)

end


                           (* +----------------+
                              |  Statistiques  |
                              +----------------+ *)

module Count(A:Int_type) = struct
  open Printf

  type statelt = {
    mutable n:float;  (* nombre d''appels,   number of calls      *)
    mutable s:float;  (* somme des tailles,  sum of operand sizes *)
    mutable m:int     (* taille maximale,    max operand size     *)
  }

  let cadd  = {n=0.0; s=0.0; m=0} (* add      sub                              *)
  let cmul  = {n=0.0; s=0.0; m=0} (* mul      sqr                              *)
  let cquo  = {n=0.0; s=0.0; m=0} (* quo      modulo       quomod              *)
  let cpow  = {n=0.0; s=0.0; m=0} (* pow      powmod       fact                *)
  let croot = {n=0.0; s=0.0; m=0} (* sqrt     root                             *)
  let cgcd  = {n=0.0; s=0.0; m=0} (* gcd      gcd_ex       cfrac       isprime *)
  let cbin  = {n=0.0; s=0.0; m=0} (* shr      shl          split       join    *)
                                  (* nbits    lowbits      highbits    nth_bit *)
                                  (* nth_word random                           *)
  let cmisc = {n=0.0; s=0.0; m=0} (* abs      neg          make_ref    copy_in *)
                                  (* copy_out comparaisons conversions         *)

  let clear_stats() =
    cadd.n  <- 0.0;  cadd.s  <- 0.0;  cadd.m  <- 0;
    cmul.n  <- 0.0;  cmul.s  <- 0.0;  cmul.m  <- 0;
    cquo.n  <- 0.0;  cquo.s  <- 0.0;  cquo.m  <- 0;
    cpow.n  <- 0.0;  cpow.s  <- 0.0;  cpow.m  <- 0;
    croot.n <- 0.0;  croot.s <- 0.0;  croot.m <- 0;
    cgcd.n  <- 0.0;  cgcd.s  <- 0.0;  cgcd.m  <- 0;
    cbin.n  <- 0.0;  cbin.s  <- 0.0;  cbin.m  <- 0;
    cmisc.n <- 0.0;  cmisc.s <- 0.0;  cmisc.m <- 0
    
  let print_stat nom c =
    if c.n = 0.0
    then printf "%5s %10d %10s %10s\n" nom 0 "-" "-"
    else printf "%5s %10d %10d %10d\n" nom (truncate c.n)
                                           (truncate(c.s/.c.n)) c.m
  let print_stats() =
    printf "%5s %10s %10s %10s\n" "op" "count" "avg.size" "max.size";
    print_stat "add"  cadd;
    print_stat "mul"  cmul;
    print_stat "quo"  cquo;
    print_stat "pow"  cpow;
    print_stat "root" croot;
    print_stat "gcd"  cgcd;
    print_stat "bin"  cbin;
    print_stat "misc" cmisc;
    flush stdout

  (*
    fonctions de comptage : signature fonction compteur
    abrviations pour les types des arguments
    tref       = r
    t          = t
    int        = i
    unit       = u
    round_mode = m
    string     = s
    bool       = b
    tristate   = c
  *)

  let i f c = fun i1 ->
    c.n <- c.n +. 1.0;
    f i1

  let i_i f c = fun i1 i2 ->
    c.n <- c.n +. 1.0;
    f i1 i2

  let m_r_r_t_t f c = fun m r1 r2 t1 t2 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f m r1 r2 t1 t2 

  let m_r_t f c = fun m r1 t1 ->
    let n1 = A.nbits(t1) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1);
    c.m <- max c.m n1;
    f m r1 t1

  let m_r_t_i f c = fun m r1 t1 i1 ->
    let n1 = A.nbits(t1) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1);
    c.m <- max c.m n1;
    f m r1 t1 i1

  let m_r_t_t f c = fun m r1 t1 t2 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f m r1 t1 t2 

  let m_r_t_t_t f c = fun m r1 t1 t2 t3 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) and n3 = A.nbits(t3) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2+n3)/.3.0;
    c.m <- max c.m (max n1 (max n2 n3));
    f m r1 t1 t2 t3

  let m_t f c = fun m t1 ->
    let n1 = A.nbits(t1) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1);
    c.m <- max c.m n1;
    f m t1

  let m_t_i f c = fun m t1 i1 ->
    let n1 = A.nbits(t1) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1);
    c.m <- max c.m n1;
    f m t1 i1

  let m_t_t f c = fun m t1 t2 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f m t1 t2

  let m_t_t_t f c = fun m t1 t2 t3 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) and n3 = A.nbits(t3) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2+n3)/.3.0;
    c.m <- max c.m (max n1 (max n2 n3));
    f m t1 t2 t3

  let r f c = fun r1 ->
    c.n <- c.n +. 1.0;
    f r1

  let r_i f c = fun r1 i1 ->
    c.n <- c.n +. 1.0;
    f r1 i1

  let r_i_i f c = fun r1 i1 i2 ->
    c.n <- c.n +. 1.0;
    f r1 i1 i2

  let r_r_r_r_r_t_t f c = fun r1 r2 r3 r4 r5 t1 t2 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f r1 r2 r3 r4 r5 t1 t2

  let r_r_r_t_t f c = fun r1 r2 r3 t1 t2 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f r1 r2 r3 t1 t2

  let r_r_t_i f c = fun r1 r2 t1 i1 ->
    let n1 = A.nbits(t1) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1);
    c.m <- max c.m n1;
    f r1 r2 t1 i1

  let r_r_t_t f c = fun r1 r2 t1 t2 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f r1 r2 t1 t2

  let r_s f c = fun r1 s1 ->
    c.n <- c.n +. 1.0;
    f r1 s1

  let r_t f c = fun r1 t1 ->
    let n1 = A.nbits(t1) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1);
    c.m <- max c.m n1;
    f r1 t1

  let r_t_i f c = fun r1 t1 i1 ->
    let n1 = A.nbits(t1) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1);
    c.m <- max c.m n1;
    f r1 t1 i1

  let r_t_t f c = fun r1 t1 t2 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f r1 t1 t2

  let r_t_t_i f c = fun r1 t1 t2 i2 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f r1 t1 t2 i2

  let r_t_t_t f c = fun r1 t1 t2 t3 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) and n3 = A.nbits(t3) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2+n3)/.3.0;
    c.m <- max c.m (max n1 (max n2 n3));
    f r1 t1 t2 t3

  let s f c = fun s1 ->
    c.n <- c.n +. 1.0;
    f s1

  let t f c = fun t1 ->
    let n1 = A.nbits(t1) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1);
    c.m <- max c.m n1;
    f t1

  let t_i f c = fun t1 i1 ->
    let n1 = A.nbits(t1) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1);
    c.m <- max c.m n1;
    f t1 i1

  let t_t f c = fun t1 t2 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f t1 t2

  let t_t_i f c = fun t1 t2 i1 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2)/.2.0;
    c.m <- max c.m (max n1 n2);
    f t1 t2 i1

  let t_t_t f c = fun t1 t2 t3 ->
    let n1 = A.nbits(t1) and n2 = A.nbits(t2) and n3 = A.nbits(t3) in
    c.n <- c.n +. 1.0;
    c.s <- c.s +. float(n1+n2+n3)/.3.0;
    c.m <- max c.m (max n1 (max n2 n3));
    f t1 t2 t3

  (* oprations  implmenter *)
  type t     = A.t
  type tref  = A.tref
  let name() = sprintf "Count(%s)" (A.name())
  let zero   = A.zero
  let one    = A.one
  exception Error of string

  let make_ref    = t               A.make_ref     cmisc
  let copy_in     = r_t             A.copy_in      cmisc
  let copy_out    = r               A.copy_out     cmisc
(*let look        = r               A.look                *)
  let add         = t_t             A.add          cadd
  let add_1       = t_i             A.add_1        cadd
  let add_in      = r_t_t           A.add_in       cadd
  let add_1_in    = r_t_i           A.add_1_in     cadd
  let sub         = t_t             A.sub          cadd
  let sub_1       = t_i             A.sub_1        cadd
  let sub_in      = r_t_t           A.sub_in       cadd
  let sub_1_in    = r_t_i           A.sub_1_in     cadd
  let mul         = t_t             A.mul          cmul
  let mul_1       = t_i             A.mul_1        cmul
  let mul_in      = r_t_t           A.mul_in       cmul
  let mul_1_in    = r_t_i           A.mul_1_in     cmul
  let quomod      = t_t             A.quomod       cquo
  let quo         = t_t             A.quo          cquo
  let modulo      = t_t             A.modulo       cquo
  let gquomod     = m_t_t           A.gquomod      cquo
  let gquo        = m_t_t           A.gquo         cquo
  let gmod        = m_t_t           A.gmod         cquo
  let quomod_in   = r_r_t_t         A.quomod_in    cquo
  let quo_in      = r_t_t           A.quo_in       cquo
  let mod_in      = r_t_t           A.mod_in       cquo
  let gquomod_in  = m_r_r_t_t       A.gquomod_in   cquo
  let gquo_in     = m_r_t_t         A.gquo_in      cquo
  let gmod_in     = m_r_t_t         A.gmod_in      cquo
  let quomod_1    = t_i             A.quomod_1     cquo
  let quo_1       = t_i             A.quo_1        cquo
  let mod_1       = t_i             A.mod_1        cquo
  let gquomod_1   = m_t_i           A.gquomod_1    cquo
  let gquo_1      = m_t_i           A.gquo_1       cquo
  let gmod_1      = m_t_i           A.gmod_1       cquo
  let quomod_1_in = r_t_i           A.quomod_1_in  cquo
  let quo_1_in    = r_t_i           A.quo_1_in     cquo
  let gquomod_1_in= m_r_t_i         A.gquomod_1_in cquo
  let gquo_1_in   = m_r_t_i         A.gquo_1_in    cquo
  let abs         = t               A.abs          cmisc
  let abs_in      = r_t             A.abs_in       cmisc
  let neg         = t               A.neg          cmisc
  let neg_in      = r_t             A.neg_in       cmisc
  let sqr         = t               A.sqr          cmul
  let pow         = t_i             A.pow          cpow
  let pow_1       = i_i             A.pow_1        cpow
  let powmod      = t_t_t           A.powmod       cpow
  let gpowmod     = m_t_t_t         A.gpowmod      cpow
  let sqr_in      = r_t             A.sqr_in       cmul
  let pow_in      = r_t_i           A.pow_in       cpow
  let pow_1_in    = r_i_i           A.pow_1_in     cpow
  let powmod_in   = r_t_t_t         A.powmod_in    cpow
  let gpowmod_in  = m_r_t_t_t       A.gpowmod_in   cpow
  let sqrt        = t               A.sqrt         croot
  let root        = t_i             A.root         croot
  let gsqrt       = m_t             A.gsqrt        croot
  let groot       = m_t_i           A.groot        croot
  let sqrt_in     = r_t             A.sqrt_in      croot
  let root_in     = r_t_i           A.root_in      croot
  let gsqrt_in    = m_r_t           A.gsqrt_in     croot
  let groot_in    = m_r_t_i         A.groot_in     croot
  let fact        = i               A.fact         cpow
  let fact_in     = r_i             A.fact_in      cpow
  let gcd         = t_t             A.gcd          cgcd
  let gcd_ex      = t_t             A.gcd_ex       cgcd
  let cfrac       = t_t             A.cfrac        cgcd
  let gcd_in      = r_t_t           A.gcd_in       cgcd
  let gcd_ex_in   = r_r_r_t_t       A.gcd_ex_in    cgcd
  let cfrac_in    = r_r_r_r_r_t_t   A.cfrac_in     cgcd
  let isprime     = t               A.isprime      cgcd
  let isprime_1   = i               A.isprime_1    cgcd
  let sgn         = t               A.sgn          cmisc
  let cmp         = t_t             A.cmp          cmisc
  let cmp_1       = t_i             A.cmp_1        cmisc
  let eq          = t_t             A.eq           cmisc
  let eq_1        = t_i             A.eq_1         cmisc
  let neq         = t_t             A.neq          cmisc
  let neq_1       = t_i             A.neq_1        cmisc
  let inf         = t_t             A.inf          cmisc
  let inf_1       = t_i             A.inf_1        cmisc
  let infeq       = t_t             A.infeq        cmisc
  let infeq_1     = t_i             A.infeq_1      cmisc
  let sup         = t_t             A.sup          cmisc
  let sup_1       = t_i             A.sup_1        cmisc
  let supeq       = t_t             A.supeq        cmisc
  let supeq_1     = t_i             A.supeq_1      cmisc
  let of_int      = i               A.of_int       cmisc
  let of_string   = s               A.of_string    cmisc
  let of_int_in   = r_i             A.of_int_in    cmisc
  let of_string_in= r_s             A.of_string_in cmisc
  let int_of      = t               A.int_of       cmisc
  let string_of   = t               A.string_of    cmisc
  let bstring_of  = t               A.bstring_of   cmisc
  let hstring_of  = t               A.hstring_of   cmisc
  let ostring_of  = t               A.ostring_of   cmisc
  let nrandom     = i               A.nrandom      cbin
  let zrandom     = i               A.zrandom      cbin
  let nrandom1    = i               A.nrandom1     cbin
  let zrandom1    = i               A.zrandom1     cbin
  let nrandom_in  = r_i             A.nrandom_in   cbin
  let zrandom_in  = r_i             A.zrandom_in   cbin
  let nrandom1_in = r_i             A.nrandom1_in  cbin
  let zrandom1_in = r_i             A.zrandom1_in  cbin
(*let random_init = i               A.random_init         *)
  let nbits       = t               A.nbits        cbin
  let lowbits     = t               A.lowbits      cbin
  let highbits    = t               A.highbits     cbin
  let nth_word    = t_i             A.nth_word     cbin
  let nth_bit     = t_i             A.nth_bit      cbin
  let shl         = t_i             A.shl          cbin
  let shr         = t_i             A.shr          cbin
  let split       = t_i             A.split        cbin
  let join        = t_t_i           A.join         cbin
  let shl_in      = r_t_i           A.shl_in       cbin
  let shr_in      = r_t_i           A.shr_in       cbin
  let split_in    = r_r_t_i         A.split_in     cbin
  let join_in     = r_t_t_i         A.join_in      cbin

  (* oprations non comptabilises *)
  let look                = A.look
  let random_init         = A.random_init       
  let toplevel_print      = A.toplevel_print
  let toplevel_print_tref = A.toplevel_print_tref
end

      (* +----------------------------------------------------------+
         |  Approximation des fonctions usuelles  valeurs relles  |
         +----------------------------------------------------------+ *)


module Rfuns(E:Int_type) = struct
  module I = Infixes(E)
  open E
  open I

  exception Error of string
  let zero_zero    = Error "0/0"
  let bit_overflow = Error "number too big"


                            (* +---------------+
                               |  Utilitaires  |
                               +---------------+ *)

  (* cration d'une fonction e_f *)
  let emake f =
    fun a p n -> if p >= 0 then f a (shl one p) n else f (shr a p) one n

  (* partie entire suprieure de i/j, i,j = entiers courts avec j > 0 *)
  let ceil i j = let x = i/j in if i mod j > 0 then x+1 else x

  (* division avec dcalage et arrondi au plus proche *)
  let ( // ) a b n =
         if n = 0 then gquo Nearest_up a b
    else if n > 0 then gquo Nearest_up (shl a n) b
    else               gquo Nearest_up a (shr b n)

  (* dcalage  droite avec arrondi au plus proche *)
  let ( >> ) a n =
    let x = shr a n in if nth_bit a (n-1) then x +. sgn(a) else x

  (* valeur absolue sans recopie inutile *)
  let abs a = if sgn(a) < 0 then neg a else a

                    (* +------------------------------+
                       |  Mmorisation de constantes  |
                       +------------------------------+ *)

  type cache_item = {
    c0 : t;                     (* valeur initiale         *)
    n0 : int;                   (* nb de bits pour c0      *)
    mutable calc : int -> unit; (* n -> approx(2^n*c)      *)
    mutable c : t;              (* valeur courante         *)
    mutable n : int;            (* nb de bits pour c       *)
    mutable d : t               (* correction pour arrondi *)
  }

  let cache_list = ref [] (* liste des valeurs mmorises *)

  (* Pour obtenir une approximation de x  p bits :

     1. Si p > x.n, recalculer une valeur plus prcise par x.calc(q) avec q >= p.
     Cette instruction modifie les champs x.c,x.n et x.d de sorte qu'on
     dispose  prsent de q bits. En particulier x.n = q.

     2. Retourner (x.c + (x.d mod 2^(k+1))) / 2^k avec k = x.n - p.
        La division doit tre effectue par dcalage sans arrondi.

     Le rle de x.d est de modifier la valeur trop prcise x.c de faon 
     retrouver celle que l'on avait enregistr la premire fois que l'on avait
     calcul au moins p bits. De la sorte, une augmentation de prcision ne
     remet pas en cause les approximations antrieures.

     Cration de valeurs  mmoriser :

     [cache f c0 n0] cre un enregistrement de type cache_item associ  la
     constante x valant initialement of_string(c0)/2^n0, puis retourne une
     fonction fx : int -> int -> E.t telle que fx a n retourne approx(a*x*2^n).
     f : int -> E.t est une fonction qui sera appele par x.calc pour augmenter
     la prcision ; l'argument est le nombre de bits voulus et le rsultat est
     l'approximation correspondante.

     [cache2 f c0 d0 n0] cre deux enregistrements de type cache_item associs aux
     constantes x,y valant initialement string_of(c0)/2^n0 et string_of(d0)/2^n0,
     puis retourne le couple (fx,fy) des fonctions associes. f : int -> E.t*E.t
     est la fonction utilise par x.calc ou y.calc pour calculer simultanment
     des nouvelles approximations de x et y ; l'argument est le nombre de bits
     voulus et le rsultat les approximations correspondantes.

  *)

  let cache,cache2 =

    (* mise  jour de x.c,x.n,x.d *)
    let update x c n =
      let p  = n - x.n in
      let d = match cmp x.c (shr c p) with
        | 1 -> one
        | _ -> zero
      in
        x.c <- c;
        x.n <- n;
        x.d <- (x.d ++ d) << p
    in

    (* recalcul avec q bits *)
    let calc  x   f q = update x (f q) q
    and calc2 x y f q = let c,d = f(q) in update x c q; update y d q
    and dummy       q = failwith "this cannot happen"
    in

    (* fonction d'interrogation *)
    let get x a n =
      if a = 0 then zero
      else begin

        (* On risque d'augmenter n de 10 + taille d'un Ocaml-int.
           Vrifie que a ne dborde pas.                          *)
        if n >= max_int - 73 then raise bit_overflow;

        (* n <- n + valuation 2-adique de a
           a <- a/2^v(a)
           l <- ceil(log_2(|a|/2^v(a)))      *)

        let rec clog l b = if b = 0 then l else clog (l+1) (b/2)   in
        let rec loop n a = if a land 1 = 0 then loop (n+1) (a/2)
                           else (clog 0 (Pervasives.abs(a)-1)),n,a in
        let l,n,a = loop n a in

          (* On a besoin de l+n bits pour x. Recalcule avec 10 bits de rserve
             si la valeur enregistre n'est pas assez prcise. *)
          if x.n < l+n then x.calc(l+n+10);

          (* arrondit la valeur cache  l+n bits puis multiplie par a *)
          if l+n < x.n - nbits(x.c) then zero
          else begin
            let p   = x.n - l - n      in
            let _,d = split x.d (p+1)  in
            let r   = shr (x.c ++ d) p in (r*.a) >> l
          end

      end
    in

      (* cache *)
      (fun f c0 n0 ->
         let c0 = of_string c0 in
         let x = {c0=c0; n0=n0; c=c0; n=n0; d=zero; calc=dummy}
         in
           x.calc <- calc x f;
           cache_list := x :: !cache_list;
           get x),

      (* cache2 *)
      (fun f c0 d0 n0 ->
         let c0 = of_string c0
         and d0 = of_string d0 in
         let x = {c0=c0; n0=n0; c=c0; n=n0; d=zero; calc=dummy}
         and y = {c0=d0; n0=n0; c=d0; n=n0; d=zero; calc=dummy}
         in
           x.calc <- calc2 x y f;
           y.calc <- calc2 x y f;
           cache_list := x :: y :: !cache_list;
           (get x),(get y))


  (* taille courante du cache *)
  let cache_bits() =
    List.fold_left (fun n x -> n + nbits(x.c)) 0 !cache_list

  (* rinitialise le cache *)
  let clear_cache() =
    List.iter (fun x -> x.c <- x.c0; x.n <- x.n0; x.d <- zero) !cache_list

                      (* +----------------------+
                         |  Calcule round(c*x)  |
                         +----------------------+ *)

  (* f = n -> approx(2^n*x) *)
  let rec round_loop f r c n p =

    if n < 0 then raise bit_overflow;

    (* calcule c*f(n) puis ajoute 1/2 (Nearest_xxx) ou 1 (Ceil)
       l'arrondi demand est la partie entire du rsultat, si elle peut
       tre dtermine.
    *)
    let x = c**f(n) in
    let x = match r with
      | Floor -> x
      | Ceil  -> x ++ (one << n)
      | _     -> x ++ (one << (n-1)) in

    let x,y = split x n in
    let y,z = match sgn(c) with 1 -> y--c, y++c | _ -> y++c, y--c in

    (* on a  x + y/2^n < res < x + z/2^n, y/2^n < 1 et z/2^n > -1

       donc si    0 <= y/2^n  et  z/2^n <= 1  alors floor = x
            si   -1 <= y/2^n  et  z/2^n <= 0  alors floor = x-1
            sinon, il faut recommencer avec plus de bits
    *)
         if (y >=. 0) & not (nth_bit (z-.1) n) then x
    else if (z <=. 0) & not (nth_bit (y+.1) n) then x-.1
    else round_loop f r c (n+p) (2*p)


  (* f = a b n -> approx(2^n*f(a/b))
     exact_f = a b -> Exact(x,y) si f(a/b) = x/y
     Irationnal si f(a/b) n'est pas rationnel

     Retourne une fonction qui calcule l'arrondi selon un mode r
     spcifi de c*f(a/b).

     On teste si a/b = 0/0 -> erreur
     puis si f(a/b) est rationnel -> gquo r (c*x) y
     puis si f(a/b) est invalide  -> erreur
     puis si c = 0 -> 0
     
     Si f(a/b) est irrationnel et c <> 0, on calcule autant de bits de f(a/b)
     que ncessaire pour dterminer l'arrondi correct.
  *)
  type result = Exact of int*int | Irrationnal | Invalid of string

  let round f exact_f r a b c =
    if (a =. 0) & (b =. 0) then raise zero_zero
    else match exact_f a b with
      | Invalid s   -> raise (Error s)
      | Exact(x,y)  -> gquo_1 r (c*.x) y
      | Irrationnal ->
          if c =. 0 then zero
          else round_loop (f a b) r c (nbits(c) + 5) 5


  (* idem pour une fonction retournant un couple *)
  type result2 = Exact2 of int*int*int*int | Irrationnal2 | Invalid2 of string

  let round2 = 

    let rec loop f r c n p =

      if n < 0 then raise bit_overflow;

      (* calcule c*f(n) puis ajoute 1/2 (Nearest_xxx) ou 1 (Ceil)
         l'arrondi demand est la partie entire du rsultat, si elle peut
         tre dtermine.
      *)
      let x1,x2 = f n          in
      let x1,x2 = c**x1, c**x2 in
      let x1,x2 = match r with
        | Floor -> x1,x2
        | Ceil  -> let u = one << n     in (x1++u),(x2++u)
        | _     -> let u = one << (n-1) in (x1++u),(x2++u)
      in
      let x1,y1 = split x1 n
      and x2,y2 = split x2 n in
      let y1,z1,y2,z2 = match sgn(c) with
        | 1 -> y1--c, y1++c, y2--c, y2++c
        | _ -> y1++c, y1--c, y2++c, y2--c in

      let x1,ok =
             if      (y1 >=. 0) & not (nth_bit (z1-.1) n) then x1,   true
        else if      (z1 <=. 0) & not (nth_bit (y1+.1) n) then x1-.1,true
        else                                                   x1,   false in
      let x2,ok =
             if ok & (y2 >=. 0) & not (nth_bit (z2-.1) n) then x2,   true
        else if ok & (z2 <=. 0) & not (nth_bit (y2+.1) n) then x2-.1,true
        else                                                   x2,   false in

        if ok then x1,x2 else loop f r c (n+p) (2*p)
    in

      fun f exact_f r a b c ->
        if (a =. 0) & (b =. 0) then raise zero_zero
        else match exact_f a b with
          | Invalid2 s          -> raise (Error s)
          | Exact2(x1,y1,x2,y2) -> (gquo_1 r (c*.x1) y1), (gquo_1 r (c*.x2) y2)
          | Irrationnal2        -> loop (f a b) r c (nbits(c) + 5) 5


            (* +----------------------------------------------+
               |  Sommation de la srie entire arctan/argth  |
               +----------------------------------------------+ *)


  (* Entre :
     s = boolen
     a,b = entiers avec 0 <= |a| < |b|/sqrt(2)
     n = nombre de bits dsirs

     Sortie :
     retourne approx(2^n*argth(a/b)) (s=false) ou approx(arctan(a/b)) (s=true)

     Algorithme de calcul :
     La somme de la srie entire est calcule terme  terme avec
     regroupement des termes 2 par 2, puis 4 par 4, etc. A un moment
     donn on dispose de :

     k : on va calculer le terme a^k/(k*b^k)
     p : nombre de termes calculs
     r : minorant de log_2(k*b^k/a^k)
     pile : (a0,b0,n0,u0,v0) :: (a1,b1,n1,u1,v1) :: ...

     avec ai = (+/-a^2)^(2^i), bi = b^(2^(i+1)),
     bi/ai >= 2^ni
     ui/vi = fraction en attente pour le groupe de 2^(i+1) termes.

     La somme des p premiers termes de la srie est gale  :

     ((( u(p0)*a(p1)   u(p1) ) a(p2)   u(p2) )        u(pj) )  a*b
     ((( ----------- + ----- )*----- + ----- )* ... + ----- )*-----
     ((( v(p0)*b(p0)   v(p1) ) b(p1)   v(p2) )        v(pj) ) b(pj)

     avec p = 2^p0 + 2^p1 + ... + 2^pj, p0 < p1 < ... < pj.

     Le reste de la srie est major en valeur absolue par 2*a^k/(k*b^k).
     On poursuit les calculs tant que ce majorant est suprieur  1/2^(n+2),
     r contient un majorant de log_2(k*bk/ak), mis  jour incrmentalement
      l'aide des ni.

     Si r < n+2, on introduit le terme 1/k dans la pile, et on la rduit
     en effectuant les additions des groupes ainsi complts.

     Si r >= n+2, on effectue les additions en instance dans la pile puis
     on tronque la fraction obtenue  n bits.
  *)

  let ps_atan s =

    (* na = nbits(a), nb = nbits(b) *)
    let compute a b na nb n =

      (* empile un nouveau terme *)
      let rec empile u v p r = function
        | [] ->
            let a0 = if s then neg(sqr a) else sqr(a) in
            let b0 = sqr(b)                 in
            let n0 = nbits(b0)-nbits(a0)-1  in
              (r+n0+1),[(a0,b0,n0,u,v)]

        | [(ai,bi,ni,ui,vi)] ->
            let aj = sqr ai                     in
            let bj = sqr bi                     in
            let nj = nbits(bj)-nbits(aj)-1      in
            let uj = ui**(v**bi) ++ u**(vi**ai) in
            let vj = v**vi                      in
              (r+nj-ni+1),[(ai,bi,ni,zero,one); (aj,bj,nj,uj,vj)]

        | (ai,bi,ni,ui,vi)::suite when p land 1 = 1 ->
            let uj = ui**(v**bi) ++ u**(vi**ai) in
            let vj = v**vi                      in
            let r,pile = empile uj vj (p/2) (r-ni) suite in
              r,((ai,bi,ni,zero,one) :: pile)

        | (ai,bi,ni,_,_)::suite -> (r+ni),((ai,bi,ni,u,v)::suite)
      in

      (* effectue les additions restant dans la pile *)
      let rec depile u v p = function
        | [] -> (u,v)

        | (ai,bi,ni,ui,vi)::suite when p land 1 = 1 ->
            depile (ui**v ++ u**(ai**vi)) ((vi**v)**bi) (p/2) suite

        | _::suite -> depile u v (p/2) suite
      in

      (* lance les calculs *)
      let rec loop k r pile = if r < n+2
      then let r,pile = empile one (of_int k) (k/2) r pile in loop (k+2) r pile
      else depile zero one (k/2) pile
      in
      let u,v = loop 1 (nb-na-1) [] in (((a**b)**u) // v) n

    in

    (* acclration : soit b = qa + r avec |r| <= |a|/2. On a :

        argth(a/b) =  argth(1/q) -  argth(r/(qb-a))
       arctan(a/b) = arctan(1/q) - arctan(r/(qb+a)) 

       Le temps de calcul de compute pour a/b est de l'ordre de

                       2*nbits(b) + nbits(a)
       M(n) * ln(n) * -----------------------
                             nbits(q)

       o M(n) est le temps, suppos quasi-linaire, d'une multiplication
       dont le rsultat tient sur n bits. Les dcompositions prcdentes sont
       donc intressantes si 7*nbits(a) > 4*nbits(b)
    *)

    let rec accel a b n =
      if (a =. 0) or (n <= 0) then zero
      else if n >= max_int - 2 then raise bit_overflow
      else begin
        let na = nbits(a) and nb = nbits(b) in
        if 7*na <= 4*nb then compute a b na nb n
        else begin
          let q,r = gquomod Nearest_up b a           in
          let x = compute one q 1 (nbits q) (n+2)    in
          let c = if s then q**b ++ a else q**b -- a in
          let y = accel r c (n+2)                    in (x -- y) >> 2
        end
      end

    in accel

                         (* +---------------------+
                            |  Calcul de ln(a/b)  |
                            +---------------------+ *)

  (* Algorithme :

     si a et b sont nuls ou de signes opposs, erreur
     sinon, changer les signes et permuter au besoin pour avoir 0 < b < a

     Dterminer p tel que 1/2 < a/(2^p*b) < 2
     Retourner la fraction au besoin pour calculer ln(x) avec 1 <= x < 2

     Comparer x aux valeurs suivantes et changer x en l'expression associe
      l'intervalle contenant x.

     1. x | 1      5/4        3/2        8/5       2  |
        --+-------------------------------------------|
          |     x       3/2x       2x/3       2/x     |
        ----------------------------------------------+


     2. x | 1     16/15       9/8        6/5      5/4 |
        --+-------------------------------------------|
          |     x       9/8x        8x/9      5/4x    |
        ----------------------------------------------+


     3. x | 1    81/80      128/125     31/30      25/24       59/56     16/15|
        --+-------------------------------------------------------------------|
          |    x     80x/81      125x/128     25/24x     24x/25     16/15x    |
        ----------------------------------------------------------------------+


     4. x | 1    161/160     81/80 |
        --+------------------------|
          |    x       81/80x      |
        ---------------------------+

     Calculer le logarithme du nombre obtenu avec ps_atan, puis reconstituer
     ln(a/b) par combinaison linaire avec ln(2),ln(3),ln(5).

     Rmq: 
     Le nombre transmis  ps_atan est infrieur ou gal  1/320
     Les transformations prcdentes dcomposent correctement les nombres
     de Hamming jusqu' 100,  l'exception de 27, 54 et 81.
  *)

  let ln =

    (* puissances de 3 et de 5 *)
    let pow3 = [|1; 3;  9;  27;  81;  243; 729; 2187; 6561; 19683; 59049; 177147|]
    and pow5 = [|1; 5; 25; 125; 625; 3125 |]
    in

    (* Calcul de ln(2), ln(3), ln(5). Formules :

       [ ln(2) ]   [ 118  171  441 ]   [ 2*argth(x0/x1) ]
       [ ln(3) ] = [ 187  271  699 ] * [ 2*argth(y0/y1) ]
       [ ln(5) ]   [ 274  397 1024 ]   [ 2*argth(z0/z1) ]

       avec
       x0 = 2*5^18 - 3^27,     x1 = 2*5^18 + 3^27,      x0/x1 ~ 1/4018
       y0 = 2^38   - 3^2*5^15, y1 = 2^38   + 3^2*5^15,  y0/y1 ~ 1/2501
       z0 = 3^8*5  - 2^15,     z1 = 3^8*5  + 2^15,      z0/z1 ~ 1/1772
    *)
    let x0 = of_string "3797046263" and x1 = of_string "15254992016237"
    and y0 = of_string  "219703819" and y1 = of_string   "549536110069"
    and z0 = of_string         "37" and z1 = of_string          "65573"
    in

    let ax = cache (ps_atan false x0 x1) "315524784848423407784245608" 100
    and ay = cache (ps_atan false y0 y1) "506805081164933785747655905" 100
    and az = cache (ps_atan false z0 z1) "715280331633428905505083056" 100
    in

      (* calcul de ln(a/b) *)
      fun a b n ->

        (* on a besoin de n+4 bits en interne *)
        if n >= max_int - 4 then raise bit_overflow;

        (* change les signes au besoin de faon  avoir a > 0, b > 0 *)
        let a,b = match sgn(a),sgn(b) with
          |  1, 1 -> a,b
          | -1,-1 -> neg(a),neg(b)
          |  0, 0 -> raise zero_zero
          |  _, _ -> raise(Error "ln")
        in

        (* classe a,b et multiplie le dnominateur par 2^p *)
        let s,a,b = match cmp a b with
          | -1 -> -1,b,a
          | _  ->  1,a,b
        in
        let p = nbits(a) - nbits(b) in
        let b = b << p              in

        (* change x en 1/x si x < 1 *)
        let s,p,a,b = match cmp a b with
          | -1 -> -s,-p, b, a
          |  _ ->  s, p, a, b
        in
          (* ici res = s*(ln(a/b) + p*ln(2)) *)

        (*
          extrait les bits de poids fort de a et b pour effectuer les comparaisons
          suivantes en simple prcision. On ne garde que 18 bits pour viter les
          dbordements.
        *)

        let aa = (highbits(a) lsr 13) land 0x3ffff
        and bb = (highbits(b) lsr 13) land 0x3ffff in

        (* invariant : res = s*(sx*ln(x) + (p+p2)*ln(2) + p3*ln(3) + p5*ln(5)) *)
        let sx, p2, p3, p5 = (1,0,0,0) in

        let (* tableau 1 *)     sx, p2,      p3,      p5,         aa,    bb =
        if  2*aa >  3*bb then
        if  5*aa >  8*bb then  -sx, p2+sx,   p3,      p5,       2*bb,    aa 
                         else   sx, p2-sx,   p3+sx,   p5,       2*aa,  3*bb
        else
        if  4*aa >  5*bb then  -sx, p2-sx,   p3+sx,   p5,       3*bb,  2*aa
                         else   sx, p2,      p3,      p5,         aa,    bb
        in

        let (* tableau 2 *)     sx, p2,      p3,      p5,         aa,    bb =
        if  8*aa >  9*bb then
        if  5*aa >  6*bb then  -sx, p2-2*sx, p3,      p5+sx,    5*bb,  4*aa 
                         else   sx, p2-3*sx, p3+2*sx, p5,       8*aa,  9*bb
        else
        if 15*aa > 16*bb then  -sx, p2-3*sx, p3+2*sx, p5,       9*bb,  8*aa
                         else   sx, p2,      p3,      p5,         aa,    bb
        in

        let (* tableau 3 *)     sx, p2,      p3,      p5,         aa,    bb =
        if 30*aa > 31*bb then
        if 24*aa > 25*bb then
        if 56*aa > 59*bb then  -sx, p2+4*sx, p3-sx,   p5-sx,   16*bb, 15*aa 
                         else   sx, p2-3*sx, p3-sx,   p5+2*sx, 24*aa, 25*bb
                         else  -sx, p2-3*sx, p3-sx,   p5+2*sx, 25*bb, 24*aa
        else
        if 80*aa > 81*bb then
        if 125*aa>128*bb then   sx, p2+7*sx, p3,      p5-3*sx,125*aa,128*bb
                         else   sx, p2-4*sx, p3+4*sx, p5-sx,   80*aa, 81*bb
                         else   sx, p2,      p3,      p5,         aa,    bb
        in

        let (* tableau 4 *)     sx, p2,      p3,      p5 =
        if aa-bb >bb/160 then  -sx, p2-4*sx, p3+4*sx, p5-sx
                         else   sx, p2,      p3,      p5
        in

        (* effectue les transformations sur a,b *)
        let u,v = if p3 < 0 then pow3.(-p3), 1   else 1,pow3.(p3)   in
        let u,v = if p5 < 0 then u*pow5.(-p5), v else u,v*pow5.(p5) in
        let u,v = if p2 < 0 then (u lsl (-p2)),v else u,(v lsl  p2) in
        let a,b = if s=1    then (a*.u, b*.v)    else (b*.v, a*.u)  in

        (* Calcul final. En sortie des rductions prcdentes, on a

           -8 <= p2 <= 12,  -11 <= p3 <= 7,  -4 <= p5 <= 5.

           Si |p| n'est pas trop grand, on peut calculer les coefficients
           de ax,ay et az en simple prcision. Sinon, on effectue ce
           calcul en multi-prcision.
        *)
        if Pervasives.abs(p) <= max_int/512 then begin
          let p2 = s*(p+p2) and p3 = s*p3 and p5 = s*p5 in
          let px = 118*p2 +  187*p3 +  274*p5
          and py = 171*p2 +  271*p3 +  397*p5
          and pz = 441*p2 +  699*p3 + 1024*p5             in
          let l  = n+4                                    in
          let lx = ps_atan false (a--b) (a++b) l          in
          (lx ++ (ax px l) ++ (ay py l) ++ (az pz l)) >> 3
        end
        else begin
          let p2 = s*p2 and p3 = s*p3 and p5 = s*p5       in
          let p  = of_int(s*p)                            in
          let px = p*.118 +. (118*p2 +  187*p3 +  274*p5)
          and py = p*.171 +. (171*p2 +  271*p3 +  397*p5)
          and pz = p*.441 +. (441*p2 +  699*p3 + 1024*p5) in
          if n >= max_int - nbits(p) - 11 then raise bit_overflow;
          let l  = n + nbits(p) + 11                      in
          let lx = ps_atan false (a--b) (a++b) l          in
          (lx ++ (ax 1 l)**px ++ (ay 1 l)**py ++ (az 1 l)**pz) >> (l-n-1)
        end

  let exact_ln a b =
         if eq a b           then Exact(0,1)
    else if sgn(a) <> sgn(b) then Invalid "ln"
    else                          Irrationnal

  let r_ln = round ln exact_ln
  let e_ln = emake ln

                (* +--------------------------------------+
                   |  Calcul de argth(a/b), arccoth(a/b)  |
                   +--------------------------------------+ *)

  (* Algorithme :

     si |a| >= |b|, erreur
     si |a/b| <= 1/128, utiliser ps_atan
     sinon, argth(a/b) <- (1/2)*ln((b+a)/(b-a))

     argcoth(a/b) = argth(b/a)
  *)

  let arctanh a b n =
    if (a =. 0) & (b =. 0)   then raise zero_zero;
    if supeq (abs a) (abs b) then raise (Error "arctanh");
    if nbits(a)-nbits(b) > 7 then ps_atan false a b n
    else if n = min_int      then zero
                             else ln (b ++ a) (b -- a) (n-1)

  let exact_arctanh a b =
         if a =. 0                then Exact(0,1)
    else if supeq (abs a) (abs b) then Invalid "arctanh"
    else                               Irrationnal

  let arccoth       a b = arctanh       b a
  let exact_arccoth a b = exact_arctanh b a

  let r_arctanh = round arctanh exact_arctanh
  let e_arctanh = emake arctanh
  let r_arccoth = round arccoth exact_arccoth
  let e_arccoth = emake arccoth


                       (* +------------------------+
                          |  Calcul de argsh(a/b)  |
                          +------------------------+ *)

  (* Algorithme :
     si b = 0, erreur
     si a = 0, argsh(a/b) <- 0

     sinon, changer les signes pour avoir a >= 0, b > 0
     k <- max(0, n+1-nbits(a))
     a <- a*2^k, b <- b*2^k
     argsh(a/b) <- +/-ln((a + sqrt(a^2+b^2))/b)
  *)
  let arcsinh a b n =
    let s,a,b = match sgn(a),sgn(b) with
      | 0, 0 -> raise zero_zero
      | _, 0 -> raise (Error "arcsinh")
      | 1, 1 ->  1,a,b
      | 1, _ -> -1,a,(neg b)
      | _, 1 -> -1,(neg a),b
      | _, _ ->  1,(neg a),(neg b)
    in
    let na = nbits(a) in
    if (a =. 0) or (n <= min_int+na-1) then zero
    else begin
        let k = max 0 (n+1-na)   in
        if k >= max_int/2 then raise bit_overflow;
        let c = sqr(a) ++ sqr(b) in
        let x = ln ((a << k) ++ (gsqrt Nearest_up (c << (2*k)))) (b << k) (n+2)
        in (if s = 1 then x else neg x) >> 2
      end

  let exact_arcsinh a b =
         if a =. 0 then Exact(0,1)
    else if b =. 0 then Invalid "arcsinh"
    else                Irrationnal

  let r_arcsinh = round arcsinh exact_arcsinh
  let e_arcsinh = emake arcsinh

                       (* +------------------------+
                          |  Calcul de argch(a/b)  |
                          +------------------------+ *)

  (* Algorithme :
     si a et b sont nuls ou a/b < 1, erreur
     sinon, changer les signes pour avoir 0 < b <= a

     k <- max(0, n+2-nbits(a))
     a <- a*2^k, b <- b*2^k
     argch(a/b) <- ln((a + sqrt(a^2-b^2))/b)
  *)
  let arccosh a b n =
    let a,b = match sgn(a),sgn(b) with
      |  1,  1 -> a,b
      | -1, -1 -> (neg a),(neg b)
      |  0,  0 -> raise zero_zero
      |  _,  _ -> raise(Error "arccosh")
    in
    let na = nbits(a) in
    match cmp a b with
      |  0 -> zero
      | -1 -> raise (Error "arccosh")
      | _ when n <= min_int+na-2 -> zero
      | _  -> let k = max 0 (n+2-na)   in
              if k >= max_int/2 then raise bit_overflow;
              let c = sqr(a) -- sqr(b) in
        (ln ((a << k) ++ (gsqrt Nearest_up (c << (2*k)))) (b << k) (n+2)) >> 2

  let exact_arccosh a b =
    let c = cmp a b in
           if c = 0       then Exact(0,1)
      else if c <> sgn(b) then Invalid "arccosh"
      else                     Irrationnal

  let r_arccosh = round arccosh exact_arccosh
  let e_arccosh = emake arccosh


                        (* +-----------------------+
                           |  Calcul de Arg(a+ib)  |
                           +-----------------------+ *)

  (* Algorithme :

     si a = b = 0, erreur
     sinon, changer les signes et permuter au besoin pour avoir 0 <= b <= a

       \_            |            _/
         \_ pi/2 + x | pi/2 - x _/
           \_        |        _/
             \_      |      _/
               \_    |    _/    x
      pi - x     \_  |  _/
      _____________\_|_/____________
                  _/ | \_
      x - pi    _/   |   \_
              _/     |     \_   -x
            _/       |       \_
          _/         |         \_
        _/ -x - pi/2 | x - pi/2  \_
      _/             |             \_


     Soient z = a+ib et z~ = a-ib. Comparer b/a aux valeurs suivantes et changer
     z en l'expression associe  l'intervalle correspondant

     1. b/a | 0       1/5         5/12         2/3        1 |
        ----+-----------------------------------------------+
            |    z       (5-i)z      (3+2i)z~     (1+i)z~   |
        ----------------------------------------------------+



     2. b/a | 0       1/18        1/11        1/7       1/5 |
        ----+-----------------------------------------------+
            |    z       (18-i)z      (7+i)z~     (5+i)z~   |
        ----------------------------------------------------+



     3. b/a | 0       1/57       1/30        3/79      1/18 |
        ----+-----------------------------------------------+
            |    z       (57-i)z     (79+3i)z~    (18+i)z~  |
        ----------------------------------------------------+



     4. b/a | 0     1/239       1/114       7/524      1/57 |
        ----+-----------------------------------------------+
            |    z      (239-i)z    (524+7i)z~    (57+i)z~  |
        ----------------------------------------------------+

     En sortie, on a 0 <= b/a < 1/217
     Calculer arctan(b/a) avec ps_atan puis reconstituer l'argument original
     par combinaison linaire avec arctan(1/2),arctan(1/3),arctan(1/5) :

     arg(  1 +  i) =   a2 +  a3
     arg(  3 + 2i) =   a2 +  a3 -  a5
     arg(  5 +  i) =               a5
     arg(  7 +  i) =   a2 -  a3
     arg( 18 +  i) =  -a2 +  a3 +  a5
     arg( 79 + 3i) = -2a2 + 3a3
     arg( 57 +  i) =   a2 - 2a3 +  a5
     arg(524 + 7i) =  2a2 -  a3 - 3a5
     arg(239 +  i) =  -a2 -  a3 + 4a5

     a2, a3, a5 sont calculs  partir d'arc-tangentes plus petits :

     [ arctan(1/2) ]   [ 26  49  7 ]   [ arctan(1149/125243) ]
     [ arctan(1/3) ] = [ 18  34  5 ] * [ arctan(1/239)       ]
     [ arctan(1/5) ]   [ 11  21  3 ]   [ arctan(369/128467)  ]

  *)

  let arg =

    (* puissances de 2+i et de 5+i (simplifies par le pgcd des parties
       relles et imaginaires). On n'a pas besoin des puissances de 3+i
       car on utilisera la relation a2 + a3 = pi/4
    *)
    let pow2 = [| (1,0);     (2,1);     (3,4);      (2,11);      (-7,24);
                  (-38,41);  (-117,44); (-278,-29); (-527,-336); (-718,-1199);
                  (-237,-3116) |]
    and pow5 = [| (1,0);     (5,1);     (12,5);     (55,37);     (119,120);
                  (475,719); (828,2035);(2105,11003) |]
    in

    (* mmorisation des arc-tangentes de base *)
    let x0 = of_int 1149 and x1 = of_int 125243
    and y0 = of_int    1 and y1 = of_int    239
    and z0 = of_int  369 and z1 = of_int 128467
    in

    let ax = cache (ps_atan true x0 x1) "11629310050144803718654399912" 100
    and ay = cache (ps_atan true y0 y1)  "5303946455430554113486419863" 100
    and az = cache (ps_atan true z0 z1)  "3641104603484748823493444480" 100
    in


      (* calcul de arg(a+ib) *)
      fun a b n ->

        (* on a besoin de n+3 bits en interne *)
        if n >= max_int - 3 then raise bit_overflow;

        if (a =.0) & (b=.0) then raise zero_zero;

        (* change les signes et permute a,b de faon  avoir 0 <= b <= a *)
        (* res = s*arg + p2*a2 + p3*a3 + p5*a5 *)
        let                s,  p2,     p3,     p5,   a, b =
        if b <. 0    then -1,  0,      0,      0,    a, neg(b)
                     else  1,  0,      0,      0,    a, b
        in

        let                s,  p2,     p3,     p5,   a, b =
        if a <. 0    then -s,  p2+4*s, p3+4*s, p5,   neg(a),b
                     else  s,  p2,     p3,     p5,   a, b
        in

        let                s,  p2,     p3,     p5,   a, b =
        if supeq b a then -s,  p2+2*s, p3+2*s, p5,   b, a
                     else  s,  p2,     p3,     p5,   a, b
        in

        (*
          extrait les bits de poids fort de a et b pour effectuer les comparaisons
          suivantes en simple prcision. On ne garde que 13 bits pour viter les
          dbordements.
        *)

        let na = nbits(a) in
        let aa = int_of(shr a (na-13))
        and bb = int_of(shr b (na-13)) in

        (* res = s*(t*arg + q2*a2 + q3*a3 + q5*a5) + p2*a2 + p3*a3 + p5*a5 *)
        let t,q2,q3,q5 = 1,0,0,0 in

        let (* tableau 1 *)      t, q2,     q3,     q5,   aa,        bb =
        if  12*bb >  5*aa  then 
        if   3*bb >  2*aa  then -t, q2+t,   q3+t,   q5,   aa+bb,     aa-bb
                           else -t, q2+t,   q3+t,   q5-t, 3*aa+2*bb, 2*aa-3*bb
        else
        if   5*bb >    aa  then  t, q2,     q3,     q5+t, 5*aa+bb,   5*bb-aa
                           else  t, q2,     q3,     q5,   aa,        bb
        in

        let (* tableau 2 *)      t, q2,     q3,     q5,   aa,        bb =
        if  11*bb >    aa  then 
        if   7*bb >    aa  then -t, q2,     q3,     q5+t, 5*aa+bb,   aa-5*bb
                           else -t, q2+t,   q3-t,   q5,   7*aa+bb,   aa-7*bb
        else
        if  18*bb >    aa  then  t, q2-t,   q3+t,   q5+t, 18*aa+bb,  18*bb-aa
                           else  t, q2,     q3,     q5,   aa,        bb
        in

        let (* tableau 3 *)      t, q2,     q3,     q5,   aa,        bb =
        if  30*bb >    aa  then 
        if  79*bb >  3*aa  then -t, q2-t,   q3+t,   q5+t, 18*aa+bb,  aa-18*bb
                           else -t, q2-2*t, q3+3*t, q5,   79*aa+3*bb,3*aa-79*bb
        else
        if  57*bb >    aa  then  t, q2+t,   q3-2*t, q5+t, 57*aa+bb,  57*bb-aa
                           else  t, q2,     q3,     q5,   aa,        bb
        in

        let (* tableau 4 *)      t, q2,     q3,     q5 =
        if bb > aa/114     then 
        if bb > (7*aa)/524 then -t, q2+t,   q3-2*t, q5+t
                           else -t, q2+2*t, q3-t,   q5-3*t
        else
        if bb > aa/239     then  t, q2-t,   q3-t,   q5+4*t
                           else  t, q2,     q3,     q5
        in

        (* combine les multiplicateurs *) 
        let u2,v2 = pow2.(Pervasives.abs(q2-q3))
        and u5,v5 = pow5.(Pervasives.abs q5)         in
        let v2    = if q2 >= q3 then v2 else -v2
        and v5    = if q5 >=  0 then v5 else -v5     in
        let u,v   = (u2*u5 - v2*v5), (u2*v5 + v2*u5) in
        let u,v   = match q3 land 7 with
          | 0 ->  u,    v
          | 1 ->  u-v,  u+v
          | 2 -> -v,    u
          | 3 ->  u+v,  v-u
          | 4 -> -u,   -v
          | 5 ->  v-u, -v-u
          | 6 ->  v,   -u
          | _ -> -u-v,  u-v
        in

        let a,b = match s with
          | 1 -> a*.u ++ b*.v, b*.u -- a*.v
          | _ -> a*.u ++ b*.v, a*.v -- b*.u
        in

        let p2 = p2 + s*q2 and p3 = p3 + s*q3 and p5 = p5 + s*q5 in
        let px = 26*p2 +  18*p3 +  11*p5
        and py = 49*p2 +  34*p3 +  21*p5
        and pz =  7*p2 +   5*p3 +   3*p5                         in
        let l  = n+3                                             in
        let at = ps_atan true b a l                              in
          (at ++ (ax px l) ++ (ay py l) ++ (az pz l)) >> (l-n)

  let exact_arg a b =
    if (a >. 0) & (b =. 0) then Exact(0,1) else Irrationnal

  let r_arg = round arg exact_arg
  let e_arg = emake arg


                (* +--------------------------------------+
                   |  Calcul de arctan(a/b), arccot(a/b)  |
                   +--------------------------------------+ *)


  (* arctan(a/b) = arg(u), arccot(a/b) = arg(v)

      a   b |    u     v
     -------+-------------
      +   + |  b+ia   a+ib
      +   - | -b-ia  -a-ib 
      -   + |  b+ia   a+ib
      -   - | -b-ia  -a-ib
  *) 

  let arctan a b n =
    if b <. 0 then arg (neg b) (neg a) n else arg b a n

  let arccot a b n =
    if b <. 0 then arg (neg a) (neg b) n else arg a b n

  let exact_arctan a b =
    if a =. 0 then Exact(0,1) else Irrationnal

  let exact_arccot a b =
    if (a >. 0) & (b =. 0) then Exact(0,1) else Irrationnal

  let r_arctan = round arctan exact_arctan
  let e_arctan = emake arctan
  let r_arccot = round arccot exact_arccot
  let e_arccot = emake arccot


                (* +--------------------------------------+
                   |  Calcul de arcsin(a/b), arccos(a/b)  |
                   +--------------------------------------+ *)

  (* Algorithme :

     Si |a| > |b| ou b = 0, erreur
     Sinon changer les signes au besoin pour avoir 0 < b <= |a|.

     k <- max(0, n+3 + nbits(a) - nbits(b^2))
     c <- ceil(sqrt((b^2 - a^2)*4^k))
     arcsin(a/b) <- arg(c, a*2^k)
  *)

  let arcsin a b n =
    let a,b = match sgn(b) with
      |  1 -> a,b
      | -1 -> (neg a),(neg b)
      |  _ -> raise (if a =. 0 then zero_zero else Error "arcsin")
    in

    if n < 0 then zero
    else if n >= max_int/2 -1 then raise bit_overflow
    else begin
      let b2 = sqr(b)       in
      let c2 = b2 -- sqr(a) in
      if c2 <. 0 then raise (Error "arcsin");
      let k = max 0 (n+3 + nbits(a) - nbits(b2)) in
      (arg (gsqrt Ceil(c2 << (2*k))) (a<<k) (n+2)) >> 2
    end

  let exact_arcsin a b =
         if a =. 0              then Exact(0,1)
    else if sup (abs a) (abs b) then Invalid "arcsin"
    else                             Irrationnal

  let r_arcsin = round arcsin exact_arcsin
  let e_arcsin = emake arcsin

  let arccos a b n =
    let a,b = match sgn(b) with
      |  1 -> a,b
      | -1 -> (neg a),(neg b)
      |  _ -> raise (if a =. 0 then zero_zero else Error "arccos")
    in
    let b2 = sqr(b)       in
    let c2 = b2 -- sqr(a) in
    if c2 <. 0 then raise (Error "arccos")
    else if n <= -2 then zero
    else if n >= max_int/2 -1 then raise bit_overflow
    else let k = max 0 (n+3 + nbits(a) - nbits(b2))
         in (arg (a<<k) (gsqrt Ceil(c2 << (2*k))) (n+2)) >> 2

  let exact_arccos a b =
         if eq a b              then Exact(0,1)
    else if sup (abs a) (abs b) then Invalid "arccos"
    else                             Irrationnal

  let r_arccos = round arccos exact_arccos
  let e_arccos = emake arccos


                 (* +-------------------------------------+
                    |  Sommation de la srie entire exp  |
                    +-------------------------------------+ *)

  (* Entre :
     a,b = entiers, |a| <= |b| 
     n = nb de bits dsirs

     Sortie : approx(2^n * (exp(a/b) - 1))

     Algorithme de calcul :
     La somme de la srie entire est calcule terme  terme avec
     regroupement des termes 2 par 2, puis 4 par 4, etc. A un moment
     donn on dispose de :

     k : on va calculer le terme a^k/(k!*b^k)
     p : nombre de termes dj calculs
     r : minorant de log_2(k!*b^k/a^k)
     pile : (a0,b0,n0,u0,v0) :: (a1,b1,n1,u1,v1) :: ...

     avec ai = a^(2^i), bi = b^(2^i),
     bi/ai >= 2^ni
     ui/vi = fraction en attente pour le groupe de 2^(i+1) termes.

     La somme des p premiers termes de la srie est gale  :

     ((( u(p0)*a(p1)         )    a(p2)            )              )     a
     ((( ----------- + u(p1) )*----------- + u(p2) )* ... + u(pj) )*-----------
     ((( v(p0)*b(p0)         ) b(p1)*v(p1)         )              ) b(pj)*v(pj)

     avec p = 2^p0 + 2^p1 + ... + 2^pj, p0 < p1 < ... < pj

     Le reste de la srie, y compris a^k/(k!*b^k), est major en valeur absolue
     par 2*a^k/(k!*b^k).

     Si 2*a^k/(k!*b^k) > 1/2^(n+1), on introduit le terme a^k/(k!*b^k) dans la
     pile, et on la rduit en effectuant les additions des groupes ainsi complts.

     Sinon, on effectue les additions en instance dans la pile puis on
     tronque la fraction obtenue  n bits.
  *)

  let ps_exp =

    (* na = nbits(a), nb = nbits(b) *)
    let compute a b na nb n =

      (* empile un nouveau terme *)
      let rec empile u v p r = function

        | [] -> (2*r),[(a,b,r,one,one)]

        | [(ai,bi,ni,ui,vi)] ->
            let aj = sqr ai                in
            let bj = sqr bi                in
            let nj = nbits(bj)-nbits(aj)-1 in
            let uj = ui**(v**bi) ++ u**ai  in
            let vj = v**vi                 in
              (r+nj+nbits(vj)-ni-nbits(vi)),[(ai,bi,ni,zero,one); (aj,bj,nj,uj,vj)]

        | (ai,bi,ni,ui,vi)::suite when p land 1 = 1 ->
            let uj = ui**(v**bi) ++ u**ai  in
            let vj = v**vi                 in
            let r,pile = empile uj vj (p/2) (r-ni-nbits(vi)+1) suite in
              r,((ai,bi,ni,zero,one) :: pile)

        | (ai,bi,ni,_,_)::suite -> (r+ni+nbits(v)-1),((ai,bi,ni,u,v)::suite)
      in

      (* effectue les additions restant dans la pile *)
      let rec depile u v p = function
        | [] -> (u,v)

        | (ai,bi,ni,ui,vi)::suite when p land 1 = 1 ->
            depile (ui**v ++ ai**u) ((vi**v)**bi) (p/2) suite

        | _::suite -> depile u v (p/2) suite
      in

      (* lance les calculs *)
      let rec loop k r pile =
        let v = of_int(k) in
          if r+nbits(v)-1 < n+2
          then let r,pile = empile one v (k-1) r pile in loop (k+1) r pile
          else depile zero one (k-1) pile
      in
      let u,v = loop 1 (nb-na-1) [] in ((a**u) // v) n

    in

    (* acclration : soit b = qa + r avec |r| <= |a|/2. On a

       exp(a/b) = exp(1/q) * exp(-r/qb)

       Le temps de calcul de compute pour a/b est de l'ordre de

                       4*nbits(b) + nbits(a)
       M(n) * ln(n) * -----------------------
                            2*nbits(q)

       o M(n) est le temps, suppos quasi-linaire, d'une multiplication
       dont le rsultat tient sur n bits. La dcomposition prcdente est
       donc intressante si 13*nbits(a) > 8*nbits(b)

       En calculant les deux exponentielles avec n+p bits, on obtient un
       rsultat correct si p >= 4 et n+p >= 0.
    *)
    let rec accel a b n =
      if (a =. 0) or (n < 0) then zero
      else if n >= max_int - 4 then raise bit_overflow
      else begin
        let na = nbits(a) and nb = nbits(b) in
          if 13*na <= 8*nb then compute a b na nb n
          else begin
            let q,r = gquomod Nearest_up b a        in
            let x = compute one q 1 (nbits q) (n+4) in
            let y = accel (neg r) (b**q) (n+4)      in
              (x ++ y ++ ((x**y) >> (n+4))) >> 4
          end
      end

    in accel

                    (* +------------------------------+
                       |  Calcul de exp(a), a entier  |
                       +------------------------------+ *)

  (* a = entier court, n = nombre de bits demands.

     Rmq : 88/61 < log_2(e) < 13/9 donc exp(a) < 2^(13a/9)  si a >= 0
                                        exp(a) < 2^(88a/61) si a < 0

     Algorithme :
     si a =  0, res <- 1
     si a =  1, res <- ps_exp(1)  + 1
     si a = -1, res <- ps_exp(-1) + 1
     si n + 88*a/61 < 0, res <- 0

     dans les autres cas, calculer exp(a)  partir de exp(1) ou exp(-1) par
     exponentiation binaire.

     cas a = 2b, b > 0 :
     calculer exp(b) avec n+p bits, n+2p >= 4 et p >= 13a/18+3

     cas a = 2b, b < 0 :
     calculer exp(b) avec n+p bits, n+2p >= 4 et p >= 44a/61+3

     cas a = 2b+1, b > 0, b = 3 mod 4 :
     calculer exp(2b+2) avec n+p bits et exp(-1) avec n+q bits,
     p >= 3, n+p+q >= 4 et q >= 13(a+1)/9 + 2

     cas a = 2b+1, b > 0, b <> 3 mod 4 :
     calculer exp(2b) avec n+p bits et exp(1) avec n+q bits,
     p >= 5, n+p+q >= 5 et q >= 13(a-1)/9 + 2

     cas a = 2b+1, b < 0, b = 0 mod 8 :
     calculer exp(2b-2) avec n+p bits et exp(1) avec n+q bits,
     p >= 5, n+p+q >= 4 et q >= 88(a-1)/61 + 4

     cas a = 2b+1, b < 0, b <> 0 mod 8 :
     calculer exp(2b+2) avec n+p bits et exp(-1) avec n+q bits,
     p >= 2, n+p+q >= 5 et q >= 88(a+1)/61 + 2

     Dans les cas impair, calculer exp(2b) en premier car cela demande
     en gnral une plus grande prcision pour exp(+/-1) que les n+q
     bits ncessaires, donc on profite de l'effet de cache.
  *)

  let exp_int =

    let f x n = (ps_exp (of_int x) one n) ++ (one << n) in
    let e1  = cache (f(1))  "3445831591435597602840181573642" 100
    and em1 = cache (f(-1))  "466342594412604466517458581545" 100
    in

    let rec loop a n =

      (* cas de base *)
           if a =  0 then one << n
      else if a =  1 then e1  1  n
      else if a = -1 then em1 1  n

      (* pour viter les dbordements dans les calculs sur n et a,
         on travaille en multiprcision *)
      else let aa = of_int a and nn = of_int n in

      if a < 0 then begin

        (* cas e^a < 1/2^n *)
        if nn*.61 ++ aa*.88 <. 0 then zero

        (* cas a pair *)
        else if a land 1 = 0 then begin
          let p = max (ceil (4-n) 2) (int_of (gquo_1 Ceil (aa*.44 +. 183) 61)) in
          if (p > 0) & (n >= max_int - 2*p) then raise bit_overflow;
          sqr(loop (a/2) (n+p)) >> (n+2*p)
        end

        (* cas a impair, b = 0 mod 8 *)
        else if a land 14 = 0 then begin
          let p = 5 in
          if n >= max_int - p then raise bit_overflow;
          let q = max (4-n-p) (int_of (gquo_1 Ceil (aa*.88 +. 156) 61)) in
          let x = loop (a-1) (n+p) in
          let y = e1 1 (n+q)       in
          (x**y) >> (n+p+q)
        end

        (* cas a impair, b <> 0 mod 8 *)
        else begin
          let p = 2 in
          if n >= max_int - p then raise bit_overflow;
          let q = max (5-n-p) (int_of (gquo_1 Ceil (aa*.88 +. 210) 61)) in
          let x = loop (a+1) (n+p) in
          let y = em1 1 (n+q)      in
          (x**y) >> (n+p+q)
        end

      end
      else begin

        (* cas e^a < 1/2^n *)
        if nn*.9 ++ aa*.13 <. 0 then zero

        (* cas a pair *)
        else if a land 1 = 0 then begin
          if n <= min_int + 4 then raise bit_overflow;
          let p = max (ceil (4-n) 2) (int_of (gquo_1 Ceil (aa*.13 +. 54) 18)) in
          if (n >= max_int-p) or (n+p >= max_int-p) then raise bit_overflow;
          sqr(loop (a/2) (n+p)) >> (n+2*p)
        end

        (* cas a impair, b = 3 mod 4 *)
        else if a land 6 = 6 then begin
          let p = 3 in
          if a/9 >= (max_int-4)/13 then raise bit_overflow;
          let q = max (4-n-p) (int_of (gquo_1 Ceil (aa*.13 +. 31) 9)) in
          if n >= max_int-p-q then raise bit_overflow;
          let x = loop (a+1) (n+p) in
          let y = em1 1 (n+q)      in
          (x**y) >> (n+p+q)
        end

        (* cas a impair, b <> 3 mod 4 *)
        else begin
          let p = 5 in
          if a/9 >= (max_int-1)/13 then raise bit_overflow;
          let q = max (5-n-p) (int_of (gquo_1 Ceil (aa*.13 +. 5) 9)) in
          if n >= max_int-p-q then raise bit_overflow;
          let x = loop (a-1) (n+p) in
          let y = e1 1 (n+q)       in
          (x**y) >> (n+p+q)
        end
      end

    in loop


                        (* +----------------------+
                           |  Calcul de exp(a/b)  |
                           +----------------------+ *)

  (* Algorithme :

     si a = b = 0, erreur
     si a < 0 et b = 0, res <- 0

     si a/b < 0 :
     si n + (88/61)*(a/b) < 0, res <- 0
     sinon, soit a = bc + d avec |d| <= b/2
     calculer exp_int(c) avec n+p bits, ps_exp(d/b) avec n+q bits,
     puis tronquer le produit  n bits. Le rsultat est correct si
     p >= 4, n+p+q >= 5, q >= 88c/61 + 2.

     si a/b > 0 :
     soit a = bc + d avec |d| <= b/2
     calculer exp_int(c) avec n+p bits, ps_exp(d/b) avec n+q bits,
     puis tronquer le produit  n bits. Le rsultat est correct si
     p >= 4, n+p+q >= 5, q >= 13c/9 + 2.
  *)

  let exp a b n =

      if b =. 0 then match sgn(a) with 
        | 0 -> raise zero_zero
        | 1 -> raise (Error "exp")
        | _ -> zero

      else let a,b = if b >. 0 then a,b else (neg a),(neg b) in

      if a <. 0
      then if a*.88 ++ (b*.61)*.n <. 0 then zero
      else begin

        if n >= max_int - 6 then raise bit_overflow;
        let p = 4                                    in
        let c,d = gquomod Nearest_up a b             in
        let q = max (5-n-p) (int_of (gquo_1 Ceil (c*.88 +. 122) 61)) in
        let x = exp_int (int_of c) (n+p)             in
        let y = ps_exp d b (n+q)                     in
        (x ++ (x << (n+q)) ++ x**y) >> (n+p+q)

      end

      else if a*.13 ++ (b*.9)*.n <. 0 then zero
      else begin

        let c,d = gquomod Nearest_up a b             in
        let q = gquo_1 Ceil (c*.13 +. 18) 9          in
        if q >=. max_int then raise bit_overflow;
        let p = 4                                    in
        let q = max (5-n-p) (int_of q)               in
        if n >= max_int-p-q then raise bit_overflow;
        let x = exp_int (int_of c) (n+p)             in
        let y = ps_exp d b (n+q)                     in
        (x ++ (x << (n+q)) ++ x**y) >> (n+p+q)

      end

  let exact_exp a b =
         if a =. 0              then Exact(1,1)
    else if (a <. 0) & (b =. 0) then Exact(0,1)
    else                             Irrationnal

  let r_exp = round exp exact_exp
  let e_exp = emake exp

                    (* +------------------------------+
                       |  Calcul de ch(a/b), sh(a/b)  |
                       +------------------------------+ *)

  let cosh a b n =
    if n >= max_int-1 then raise bit_overflow;
    ((exp a b (n+1)) ++ (exp (neg a) b (n+1))) >> 2

  let sinh a b n =
    if n >= max_int-1 then raise bit_overflow;
    ((exp a b (n+1)) -- (exp (neg a) b (n+1))) >> 2

  let cosinh a b n =
    if n >= max_int-1 then raise bit_overflow;
    let x = exp a       b (n+1)
    and y = exp (neg a) b (n+1) in (x++y) >> 2, (x--y) >> 2

  let exact_cosinh a b =
    if a =. 0 then Exact2(1,1,0,1) else Irrationnal2

  let exact_cosh a b =
    if a =. 0 then Exact(1,1) else Irrationnal

  let exact_sinh a b =
    if a =. 0 then Exact(0,1) else Irrationnal

  let r_cosinh = round2 cosinh exact_cosinh
  let e_cosinh = emake  cosinh
  let r_cosh   = round  cosh   exact_cosh
  let e_cosh   = emake  cosh
  let r_sinh   = round  sinh   exact_sinh
  let e_sinh   = emake  sinh


                         (* +---------------------+
                            |  Calcul de th(a/b)  |
                            +---------------------+ *)

  (* Algorithme :

     si a = b = 0, erreur sinon si b=0, th(a/b) <- sgn(a)
     si n < 0, th(a/b) <- 0

     sinon, changer les signes au besoin pour avoir a >= 0, b > 0
     th(a/b) = (1 - exp(-2a/b)) / (1 + exp(-2a/b))

     Calculer exp(-2a/b) avec n+p bits, le quotient arrondi est correct si
     p >= 2.
  *)
  let tanh a b n =

    if b =. 0 then begin
      if a =. 0 then raise zero_zero
      else of_int(sgn a) << n
    end
    else if (a =. 0) or (n < 0) then zero

    else begin

      let s,a,b = match sgn(a),sgn(b) with
        | -1, -1 -> false, (neg a), (neg b)
        | -1,  _ -> true,  (neg a), b
        |  _, -1 -> true,  a,       (neg b)
        |  _,  _ -> false, a,       b
      in

      if n >= max_int-2 then raise bit_overflow;
      let x = exp (a*.(-2)) b (n+2) in
      let y = one << (n+2)          in
      ((if s then x--y else y--x) // (x ++ y)) n

    end

  let exact_tanh a b =
         if a =. 0 then Exact(0,1)
    else if b =. 0 then Exact(sgn(a),1)
    else                Irrationnal

  let r_tanh = round tanh exact_tanh
  let e_tanh = emake tanh


                        (* +-----------------------+
                           |  Calcul de coth(a/b)  |
                           +-----------------------+ *)


  (* Algorithme :

     si a = 0, erreur sinon si b = 0, coth(a/b) <- sgn(a)

     sinon, changer les signes au besoin pour avoir a > 0, b > 0
     coth(a/b) = (1 + exp(-2a/b)) / (1 - exp(-2a/b))

     Calculer exp(-2a/b) avec p bits, le quotient arrondi est correct si
     p - n + 2k >= 4 et p + k >= 3, avec k = min(0, nbits(a) - nbits(b))

  *)
  let coth a b n =

    if a =. 0 then raise (if b =. 0 then zero_zero else Error "coth");
    if b =. 0 then of_int(sgn a) << n

    else begin

      let s,a,b = match sgn(a),sgn(b) with
        | -1, -1 -> false, (neg a), (neg b)
        | -1,  _ -> true,  (neg a), b
        |  _, -1 -> true,  a,       (neg b)
        |  _,  _ -> false, a,       b
      in

      let k = min 0 (nbits(a) - nbits(b)) in
      let q = 4 - 2*k                     in
      if n >= max_int - q then raise bit_overflow;
      let p = max (3 - k) (n + q)         in
      let x = exp (a*.(-2)) b p           in
      let y = one << p                    in
      let u = x ++ y                      in
      let v = if s then x--y else y--x    in (u // v) n
                                            
    end

  let exact_coth a b =
         if a =. 0 then Invalid "coth"
    else if b =. 0 then Exact(sgn(a),1)
    else                Irrationnal

  let r_coth = round coth exact_coth
  let e_coth = emake coth


              (* +------------------------------------------+
                 |  Sommation de la srie entire exp(i*x)  |
                 +------------------------------------------+ *)

  (* Algorithme identique  ps_exp, adapt au cas complexe. Changer u en u + it
     dans les explications. *)

  let ps_cosin =

    (* na = nbits(a), nb = nbits(b) *)
    let compute a b na nb n =

      (* empile un nouveau terme *)
      let rec empile u t v first p r = function

        | [] -> (2*r),[(a,b,r,one,zero,one)]

        | [(ai,bi,ni,ui,ti,vi)] ->
            let aj = if first then neg(sqr ai) else sqr ai in
            let bj = sqr bi                in
            let nj = nbits(bj)-nbits(aj)-1 in
            let vbi= v**bi                 in
            let uj = ui**vbi ++ u**ai      in
            let tj = ti**vbi ++ t**ai      in
            let vj = v**vi                 in
              (r+nj+nbits(vj)-ni-nbits(vi)),
              [(ai,bi,ni,zero,zero,one); (aj,bj,nj,uj,tj,vj)]

        | (ai,bi,ni,ui,ti,vi)::suite when p land 1 = 1 ->
            let vbi= v**bi                 in
            let uj = ui**vbi ++ u**ai      in
            let tj = ti**vbi ++ t**ai      in
            let vj = v**vi                 in
            let r,pile = empile uj tj vj false (p/2) (r-ni-nbits(vi)+1) suite in
              r,((ai,bi,ni,zero,zero,one) :: pile)

        | (ai,bi,ni,_,_,_)::suite -> if first
          then (r+ni+nbits(v)-1),((ai,bi,ni,t,u,v)::suite)
          else (r+ni+nbits(v)-1),((ai,bi,ni,u,t,v)::suite)
      in

      (* effectue les additions restant dans la pile *)
      let rec depile u t v p = function
        | [] -> (neg(t),u,v)

        | (ai,bi,ni,ui,ti,vi)::suite when p land 1 = 1 ->
            depile (ui**v ++ ai**u) (ti**v ++ ai**t) ((vi**v)**bi) (p/2) suite

        | _::suite -> depile u t v (p/2) suite
      in

      (* lance les calculs *)
      let rec loop k r pile =
        let v = of_int(k) in
          if r+nbits(v)-1 < n+2
          then let r,pile = empile zero one v true (k-1) r pile in loop (k+1) r pile
          else depile zero zero one (k-1) pile
      in

      let u,t,v = loop 1 (nb-na-1) [] in (((a**u) // v) n), (((a**t) // v) n)
    in

    (* acclration : le temps de calcul de compute est de l'ordre de
       
                       7*nbits(b) + 2*nbits(a)
       M(n) * ln(n) * -------------------------
                            2*nbits(q)


       donc l'acclration est payante quand 23*nbits(a) > 14*nbits(b).
       Prendre p >= 3, n+p >= 4 pour les exponentielles intermdiaires.
    *)

    let rec accel a b n =
      if (a =. 0) or (n < 0) then (zero,zero)
      else begin
        let na = nbits(a) and nb = nbits(b) in
          if 23*na <= 14*nb then compute a b na nb n
          else begin
            let q,r = gquomod Nearest_up b a            in
            let p = max 3 (4-n)                         in
            if n >= max_int - p then raise bit_overflow;
            let x1,x2 = compute one q 1 (nbits q) (n+p) in
            let y1,y2 = accel (neg r) (b**q) (n+p)      in
              (x1 ++ y1 ++ ((x1**y1 -- x2**y2) >> (n+p))) >> p,
              (x2 ++ y2 ++ ((x1**y2 ++ x2**y1) >> (n+p))) >> p
          end
      end

    in accel

                 (* +------------------------------------+
                    |  Calcul de exp(ia), a entier long  |
                    +------------------------------------+ *)

  (* Algorithme :
     si a =  0, res <- 1
     si n <= 0, res <- 0
     si a =  1, res <- ps_cosin(1) + 1
     si a < 0,  res <- conj(exp(-ia))

     dans les autres cas, calculer exp(ia)  partir de exp(i) par
     exponentiation binaire.

     cas a = 2b :
     calculer exp(ib) avec n+3 bits

     cas a = 2b+1 :
     calculer exp(ib) avec n+3 bits et exp(i) avec n+4 bits.
  *)

  let cosin_int =

    let c1,s1 = cache2 (fun n -> let c,s = ps_cosin one one n in c ++ (one << n),s)
                   "684914542338443705542701930871"
                  "1066691198966369310635496214684" 100 in

    (* k = rang du bit de a  examiner
       u+iv = 2^(n+3(na-k))*exp(i*(a / 2^k)) *)
    let rec loop a k u v n =

      if k < 0 then u,v

      else if nth_bit a k then begin
        let u,v = sqr(u)--sqr(v), (u**v)*.2  in
        let c,s = (c1 1 (n+1)), (s1 1 (n+1)) in
        loop a (k-1) ((c**u--s**v)>>(2*n+4)) ((c**v++s**u)>>(2*n+4)) (n-3)
      end

      else begin
        let u,v = sqr(u)--sqr(v), (u**v)*.2  in
        loop a (k-1) (u>>(n+3)) (v>>(n+3)) (n-3)
      end

    in

      (* lance les calculs *)
      fun a n ->
             if a =. 0 then (one << n),zero
        else if n <= 0 then zero,zero
        else begin
          let na = nbits(a)     in
          if n >= max_int/2 + 1 - 3*na then raise bit_overflow;
          let n =  n + 3*na - 3 in
          let u,v = loop a (na-2) (c1 1 n) (s1 1 n) n in
          if a >. 0 then u,v else u,neg(v)
        end

                        (* +-----------------------+
                           |  Calcul de exp(ia/b)  |
                           +-----------------------+ *)

  let cosin a b n =

         if b =. 0 then raise(if a =. 0 then zero_zero else Error "cosin")
    else if a =. 0 then one << n, zero
    else if n <= 0 then zero, zero
    else begin
      if n >= max_int - 6 then raise bit_overflow;
      let q,r = gquomod Nearest_up a b in
      let c,s = cosin_int q   (n+3)    in
      let u,v = ps_cosin  r b (n+3)    in
      let u   = u ++ (one << (n+3))    in
        (c**u -- s**v)>>(n+6), (c**v ++ s**u)>>(n+6)
    end

  let cos a b n =

         if b =. 0 then raise(if a =. 0 then zero_zero else Error "cos")
    else if a =. 0 then one << n
    else if n <= 0 then zero
    else begin
      if n >= max_int - 6 then raise bit_overflow;
      let q,r = gquomod Nearest_up a b in
      let c,s = cosin_int q   (n+3)    in
      let u,v = ps_cosin  r b (n+3)    in
      let u   = u ++ (one << (n+3))    in
        (c**u -- s**v)>>(n+6)
    end

  let sin a b n =

         if b =. 0 then raise(if a =. 0 then zero_zero else Error "sin")
    else if a =. 0 then zero
    else if n <= 0 then zero
    else begin
      if n >= max_int - 6 then raise bit_overflow;
      let q,r = gquomod Nearest_up a b in
      let c,s = cosin_int q   (n+3)    in
      let u,v = ps_cosin  r b (n+3)    in
      let u   = u ++ (one << (n+3))    in
        (c**v ++ s**u)>>(n+6)
    end

  let exact_cosin a b =
    if a =. 0 then Exact2(1,1,0,1) else Irrationnal2

  let exact_cos a b =
    if a =. 0 then Exact(1,1) else Irrationnal

  let exact_sin a b =
    if a =. 0 then Exact(0,1) else Irrationnal

  let r_cosin = round2 cosin exact_cosin
  let e_cosin = emake  cosin
  let r_cos   = round  cos   exact_cos
  let e_cos   = emake  cos
  let r_sin   = round  sin   exact_sin
  let e_sin   = emake  sin

                  (* +----------------------------------+
                     |  Calcul de tan(a/b), cotan(a/b)  |
                     +----------------------------------+ *)

  (* Algorithme pour tan(a/b) :

     1. p <- 5
     2. x+iy <- approx(2^(n+p) * exp(ia/b))
     3. tant que |x| < 8 : p <- 2*p, recalculer x,y
     4. soit r tel que 2^r < |x| <= 2^(r+1)
        k <- max( n+3-r, 2n+4+p-2r )
        tant que k > 0 : p <- p+k, recalculer x,y,r,k
     5. res <- approx(2^n*x/y)

     Pour cotan(a/b) : idem en permutant x,y
  *)

  let tan,cot =

    let rec loop f n p =
      if n >= (max_int - p - 4)/2 then raise bit_overflow;
      let x,y = f (n+p) in
      let nx = nbits(x) in
      if nx < 4 then begin
        if p >= max_int/2 then raise bit_overflow;
        loop f n (2*p)
      end
      else begin
        let r = let _,t = split x (nx-1) in if t =. 0 then nx-2 else nx-1 in
        let k = max (n+3-r) (2*n+4+p-2*r) in
        if k > 0 then loop f n (p+k) else (y // x) n
      end
    in

    (* tan *)
    (fun a b n ->
       match sgn(a),sgn(b) with
         | 0, 0 -> raise zero_zero
         | _, 0 -> raise (Error "tan")
         | 0, _ -> zero
         | _, _ -> loop (cosin a b) n 5),

    (* cot *)
    (fun a b n ->
       match sgn(a),sgn(b) with
         | 0, 0 -> raise zero_zero
         | _, 0 -> raise (Error "cot")
         | 0, _ -> raise (Error "cot")
         | _, _ -> loop (fun m -> let x,y = cosin a b m in y,x) n 5)
        
  let exact_tan a b =
         if a =. 0 then Exact(0,1)
    else if b =. 0 then Invalid "tan"
    else                Irrationnal

  let exact_cot a b =
    if (a =. 0) or (b =. 0) then Invalid "cot" else Irrationnal

  let r_tan = round tan exact_tan
  let e_tan = emake tan
  let r_cot = round cot exact_cot
  let e_cot = emake cot


                    (* +------------------------------+
                       |  Accs externe  round_loop  |
                       +------------------------------+ *)

  let round f r c = round_loop f r c (nbits(c) + 5) 5

end

                      (* +---------------------------+
                         |  Slection  l'excution  |
                         +---------------------------+ *)

module type Main_type = sig
  val main : string list -> unit
end

module Start(Main : functor(E:Int_type) -> Main_type) = struct
  open Printf

  let module_list =
     "clong" :: 
  (* "dlong" :: *)
     "slong" :: 
     "big"   :: 
     "gmp"   :: 
  []

  let error msg  =
    printf "%s\n" msg;
    printf "available modules :";
    List.iter (fun x -> printf " %s" x) module_list;
    printf "\n";
    flush   stdout;
    exit(1)

  (* analyse la ligne de commande et retire les options -e xxx et -count *)
  let rec parse mlist count opts = function
    | "-count"::s-> parse mlist true opts s
    | "-e"::e::s -> parse (e::mlist) count opts s
    | "-e"::_    -> error "unspecified integers"
    | o::s       -> parse mlist count (o::opts) s
    | []         -> mlist,count,(List.rev opts)

  (* lancement d'un module Main(Cmp(X,Y)) ou Main(Count(Cmp(X,Y))) *)
  module Select(A:Int_type) = struct
    let start mlist count opts = match count,mlist with
    | _,    []         -> failwith "this can't happen"
    | false,"clong"::_ -> let module M = Main(Cmp(Clong)(A))  in M.main opts 
  (*| false,"dlong"::_ -> let module M = Main(Cmp(Dlong)(A))  in M.main opts *)
    | false,"slong"::_ -> let module M = Main(Cmp(Slong)(A))  in M.main opts 
    | false,"big"  ::_ -> let module M = Main(Cmp(Big)(A))    in M.main opts 
    | false,"gmp"  ::_ -> let module M = Main(Cmp(Gmp)(A))    in M.main opts 
    | true, "clong"::_ -> let module M = Count(Cmp(Clong)(A)) in let module N = Main(M) in N.main opts; M.print_stats() 
  (*| true, "dlong"::_ -> let module M = Count(Cmp(Dlong)(A)) in let module N = Main(M) in N.main opts; M.print_stats() *)
    | true, "slong"::_ -> let module M = Count(Cmp(Slong)(A)) in let module N = Main(M) in N.main opts; M.print_stats() 
    | true, "big"  ::_ -> let module M = Count(Cmp(Big)(A))   in let module N = Main(M) in N.main opts; M.print_stats() 
    | true, "gmp"  ::_ -> let module M = Count(Cmp(Gmp)(A))   in let module N = Main(M) in N.main opts; M.print_stats() 
    | _,    m::_       -> error ("unknown integers: " ^ m)
  end

  let start() =
    let mlist,count,opts = parse [] false [] (Array.to_list Sys.argv) in
    let mlist = match mlist,module_list with
      | [],(m::_) -> [m]
      | _         -> mlist
    in
    match count,mlist with
      | _,    []         -> failwith "this can't happen"
      | false,["clong"]  -> let module M = Main(Clong)   in M.main opts 
    (*| false,["dlong"]  -> let module M = Main(Dlong)   in M.main opts *)
      | false,["slong"]  -> let module M = Main(Slong)   in M.main opts 
      | false,["big"  ]  -> let module M = Main(Big)     in M.main opts 
      | false,["gmp"  ]  -> let module M = Main(Gmp)     in M.main opts 
      | true, ["clong"]  -> let module M = Count(Clong)  in let module N = Main(M) in N.main opts; M.print_stats() 
    (*| true, ["dlong"]  -> let module M = Count(Dlong)  in let module N = Main(M) in N.main opts; M.print_stats() *)
      | true, ["slong"]  -> let module M = Count(Slong)  in let module N = Main(M) in N.main opts; M.print_stats() 
      | true, ["big"  ]  -> let module M = Count(Big)    in let module N = Main(M) in N.main opts; M.print_stats() 
      | true, ["gmp"  ]  -> let module M = Count(Gmp)    in let module N = Main(M) in N.main opts; M.print_stats() 
      | _,    "clong"::t -> let module M = Select(Clong) in M.start t count opts 
    (*| _,    "dlong"::t -> let module M = Select(Dlong) in M.start t count opts *)
      | _,    "slong"::t -> let module M = Select(Slong) in M.start t count opts 
      | _,    "big"  ::t -> let module M = Select(Big)   in M.start t count opts 
      | _,    "gmp"  ::t -> let module M = Select(Gmp)   in M.start t count opts 
      | _,    m::_       -> error ("unknown integers: " ^ m)

end
